from copy import deepcopy
from .process import replace_dashes, replace_whitespace, normalize, \
    get_capitalization_pattern

__all__ = [
    "Match",
    "generate_match",
    "score_string_match",
    "score_status",
    "score",
]


class Match(object):
    """Class representing a match between a query and a reference string"""
    def __init__(self, query, ref, exact=None, space_mismatch=None,
                 dash_mismatches=None, cap_combos=None):
        self.query = query
        self.ref = ref
        self.exact = exact if exact is not None else False
        self.space_mismatch = space_mismatch if space_mismatch is not None \
            else False
        self.dash_mismatches = dash_mismatches if dash_mismatches is not None \
            else {}
        self.cap_combos = cap_combos if cap_combos is not None else []

    def __str__(self):
        return 'Match(%s)' % (','.join(['%s=%s' % (k, v) for k, v in
                                        self.__dict__.items()]))

    def __repr__(self):
        return str(self)

    def to_json(self):
        return {
            'query': self.query,
            'ref': self.ref,
            'exact': self.exact,
            'space_mismatch': self.space_mismatch,
            'dash_mismatches': list(self.dash_mismatches),
            'cap_combos': self.cap_combos
        }

    def _query_cases(self):
        return {c[0] for c in self.cap_combos}

    def _ref_cases(self):
        return {c[1] for c in self.cap_combos}

    def score_short_abbr(self):
        if len(self.ref) <= 3 and \
                (('all_caps', 'all_lower') in self.cap_combos or
                 ('all_lower', 'all_caps') in self.cap_combos):
            return 0
        else:
            return 1

    def score_mixed(self):
        if ('mixed', 'mixed') in self.cap_combos:
            return 0
        elif ('mixed' in self._query_cases()) or \
                ('mixed' in self._ref_cases()):
            return 1
        else:
            return 2

    def score_exact(self):
        return 1 if self.exact is True else 0

    def score_acic(self):
        if self.exact is True and not self.cap_combos:
            return 2
        elif set(self.cap_combos) == {('sentence_initial', 'all_lower')}:
            return 2
        if self.exact is True and set(self.cap_combos) <= \
                {('all_caps', 'sentence_initial_cap'),
                 ('sentence_initial_cap', 'all_caps')}:
            return 1
        else:
            return 0

    def score_combo(self):
        qc = self._query_cases()
        rc = self._ref_cases()
        query_combo = 4 - len(qc)
        ref_combo = 4 - len(rc)
        if 'single_cap_letter' in qc and \
                ('all_caps' in qc or 'initial_cap' in qc):
            query_combo += 1
        if 'single_cap_letter' in rc and \
                ('all_caps' in rc or 'initial_cap' in rc):
            ref_combo += 1
        if 'sentence_initial_cap' in qc and \
                (len(qc) == 1 and
                 ('sentence_initial_cap', 'all_lower') in self.cap_combos) \
                or \
                 {'single_cap_letter', 'initial_cap', 'all_lower'} & qc:
            query_combo += 1
        combo = max(query_combo, ref_combo)
        return combo

    def score_dash(self):
        return 2 - len(self.dash_mismatches)


def generate_match(query, ref, beginning_of_sentence=False):
    """Return a match data structure based on comparing a query to a ref str.

    Parameters
    ----------
    query : str
        The string to be compared against a reference string.
    ref : str
        The reference string against which the incoming query string is
        compared.
    beginning_of_sentence : bool
        True if the query_str appears at the beginning of a sentence, relevant
        for how capitalization is evaluated.

    Returns
    -------
    Match
        A Match object characterizing the match between the two strings.
    """
    # Pre-process both strings first by replacing multiple white spaces
    # with a single ASCII space, and all kinds of dashes with a single
    # ASCII dash.
    query = replace_dashes(replace_whitespace(query))
    # Corner case: some synonyms may have a trailing space which we can
    # safely remove to avoid issues in the comparison
    ref = replace_whitespace(ref).rstrip(' ')
    ref = replace_dashes(ref)

    # If we have an exact match at this point then we can return immediately
    if not beginning_of_sentence and query == ref:
        return Match(query, ref, exact=True)
    query_suffix = query
    ref_suffix = ref
    query_pieces = ['']
    ref_pieces = ['']
    dash_mismatches = set()
    while query_suffix and ref_suffix:
        # Deal with spaces first
        qs = (query_suffix[0] == ' ')
        rs = (ref_suffix[0] == ' ')
        # If both have spaces, we start new pieces and skip the spaces
        if qs and rs:
            query_suffix = query_suffix[1:]
            ref_suffix = ref_suffix[1:]
            query_pieces.append('')
            ref_pieces.append('')
        # This means that there is a space inconsistency which we don't allow
        # and return immediately
        elif qs and not rs or rs and not qs:
            return Match(query, ref, space_mismatch=True)

        # We next deal with dashes
        qd = (query_suffix[0] == '-')
        rd = (ref_suffix[0] == '-')
        # If both are dashes, we skip them
        if qd and rd:
            query_suffix = query_suffix[1:]
            ref_suffix = ref_suffix[1:]
            query_pieces.append('')
            ref_pieces.append('')
        # If there is a mismatch, we introduce new pieces but only skip the one
        # dash and record the inconsistency
        elif qd and not rd:
            dash_mismatches.add('query')
            query_suffix = query_suffix[1:]
            query_pieces.append('')
            ref_pieces.append('')
        elif not qd and rd:
            dash_mismatches.add('ref')
            ref_suffix = ref_suffix[1:]
            query_pieces.append('')
            ref_pieces.append('')
        # Otherwise both strings start with a non space/dash character that we
        # add to the latest piece
        else:
            query_pieces[-1] += query_suffix[0]
            ref_pieces[-1] += ref_suffix[0]
            ref_suffix = ref_suffix[1:]
            query_suffix = query_suffix[1:]

    # Now that we have the final pieces in place, we can count the matches and
    # capitalization relationships
    combinations = []
    first = True
    exact = False
    for qp, rp in zip(query_pieces, ref_pieces):
        first_bos = first and beginning_of_sentence
        first = False
        if qp == rp and not first_bos:
            exact = True
        else:
            qcp = get_capitalization_pattern(qp, first_bos)
            rcp = get_capitalization_pattern(rp, False)
            if qcp == rcp and qp == rp:
                exact = True
            else:
                combinations.append((qcp, rcp))
    return Match(query, ref, dash_mismatches=dash_mismatches,
                 exact=exact, cap_combos=combinations)


def score_string_match(match):
    """Return a score between 0 and 1 for the goodness of a match.

    This score is purely based on the relationship of the two strings and
    does not take the status of the reference into account.

    Parameters
    ----------
    match : gilda.scorer.Match
        The Match object characterizing the relationship of the query and
        reference strings.

    Returns
    -------
    float
        A match score between 0 and 1.
    """
    terms = [
        (match.score_short_abbr, 2),
        (match.score_mixed, 3),
        (match.score_exact, 2),
        (match.score_acic, 3),
        (match.score_combo, 5),
        (match.score_dash, 3)
    ]
    score = 0
    norm = 1
    for fun, coeff in terms:
        score = coeff * score + fun()
        norm *= coeff
    score /= (norm - 1)
    return score


def score_status(term):
    scores = {
        'curated': 4,
        'name': 3,
        'synonym': 2,
        'former_name': 1,
    }
    return scores[term.status]


def score(match, term):
    string_match_score = score_string_match(match)
    status_score = score_status(term)
    score = ((0 * 5 + status_score) * 2 + string_match_score) / 9
    return score



import time
import requests

url_prefix = 'https://www.ncbi.nlm.nih.gov/research/pubtator-api/annotations/annotate/'
submit_url = url_prefix + 'submit/'
retrieve_url = url_prefix + 'retrieve/'


def process_file(filename):
    begin = time.time()
    data = open(filename, 'rb').read()
    res = requests.post(submit_url + 'Gene', data=data)
    session_key = res.text
    delay = 10
    res = requests.get(retrieve_url + session_key)
    while res.status_code != 200:
        res = requests.get(retrieve_url + session_key)
        print('Waiting %.2fs' % delay)
        time.sleep(delay)
    end = time.time()
    print('Overall time: %s' % (end - begin))
    return end - begin


if __name__ == '__main__':
    nrep = 10
    times = []
    for i in range(nrep):
        print('Starting iteration %d' % i)
        tt = process_file('pubtator_input.json')
        times.append(tt)
    print('Average time: %s' % (sum(times) / nrep))


import pandas as pd

from gilda import ground_df

TEST_COLUMNS = ['idx', 'text']
TEST_ROWS = [
    (1, "kras"),
    (2, "apoptosis"),
]
TEST_SOLUTIONS = ['hgnc:6407', 'go:0006915']


def test_pandas_grounding():
    df = pd.DataFrame(TEST_ROWS, columns=TEST_COLUMNS)
    ground_df(df, 'text')
    assert 'text_grounded' in df.columns
    assert list(df['text_grounded']) == TEST_SOLUTIONS

    ground_df(df, 'text', target_column="target_column")
    assert 'target_column' in df.columns
    assert list(df['target_column']) == TEST_SOLUTIONS



__version__ = '1.5.0'

import logging

from .api import get_grounder, get_models, get_names, ground, make_grounder, \
    annotate
from .grounder import Grounder, ScoredMatch, Annotation
from .pandas_utils import ground_df, ground_df_map
from .term import Term, dump_terms

__all__ = [
    'ground',
    'annotate',
    'get_models',
    'get_names',
    'get_grounder',
    'make_grounder',
    "dump_terms",
    # Classes
    'Term',
    'Grounder',
    'ScoredMatch',
    'Annotation',
    # Meta
    '__version__',
    # Pandas utilities
    'ground_df',
    'ground_df_map',
]

logging.basicConfig(format=('%(levelname)s: [%(asctime)s] %(name)s'
                            ' - %(message)s'),
                    level=logging.INFO, datefmt='%Y-%m-%d %H:%M:%S')

logger = logging.getLogger('gilda')



# -*- coding: utf-8 -*-

"""This script benchmarks the responsivenes (i.e., speed) of Gilda
on the BioCreative VII BioID corpus in three settings: when used as a Python
package, as a local web service, and when using the remote public web
service."""

import pathlib
import random
import time
from typing import Optional
from itertools import zip_longest

import click
import matplotlib.pyplot as plt
import pandas as pd
import requests
import seaborn as sns
from more_click import force_option, verbose_option
from tqdm import tqdm, trange
from tqdm.contrib.logging import logging_redirect_tqdm

import gilda
from bioid_evaluation import BioIDBenchmarker
from gilda.api import grounder

HERE = pathlib.Path(__file__).parent.resolve()
RESULTS = HERE.joinpath("results")
RESULTS.mkdir(exist_ok=True, parents=True)

RESULTS_PATH = RESULTS.joinpath("bioid_responsiveness.tsv")
RESULTS_AGG_PATH = RESULTS.joinpath("bioid_responsiveness_aggregated.tsv")
RESULTS_AGG_TEX_PATH = RESULTS.joinpath("bioid_responsiveness_aggregated.tex")
FIG_PATH = RESULTS.joinpath("bioid_responsiveness.svg")
FIG_PDF_PATH = RESULTS.joinpath("bioid_responsiveness.pdf")


def ground_package(text, **_kwargs):
    return gilda.ground(text)


def ground_package_context(text, context):
    return gilda.ground(text, context=context)


def ground_app_local_multi(batch):
    return requests.post(
            "http://localhost:8001/ground_multi",
            json=[{'text': text} for text, _ in batch]).json()


def ground_app_local_multi_context(batch):
    return requests.post(
            "http://localhost:8001/ground_multi",
            json=[{'text': text, 'context': context}
                   for text, context in batch]).json()


def ground_app_local(text, **_kwargs):
    return requests.post("http://localhost:8001/ground", json={"text": text}).json()


def ground_app_local_context(text, context):
    return requests.post(
        "http://localhost:8001/ground", json={"text": text, "context": context}
    ).json()

def ground_app_remote(text, **_kwargs):
    return requests.post(
        "http://grounding.indra.bio/ground", json={"text": text}
    ).json()


def ground_app_remote_context(text, context):
    return requests.post(
        "http://grounding.indra.bio/ground", json={"text": text, "context": context}
    ).json()


#: A list of benchmarks to run with three columns:
#:  type, uses context, function
FUNCTIONS = [
    ("Python package", False, ground_package, None),
    ("Python package", True, ground_package_context, None),
    ("Local web app", False, ground_app_local, None),
    ("Local web app", True, ground_app_local_context, None),
    ("Local web app multi", False, ground_app_local_multi, 100),
    ("Local web app multi", True, ground_app_local_multi_context, 100),
    ("Public web app", False, ground_app_remote, None),
    ("Public web app", True, ground_app_remote_context, None),
]


def run_trial(
    *, trials, corpus, func, desc: Optional[str] = None, chunk: Optional[int] = None,
    batching: Optional[int] = None
):
    rv = []
    outer_it = trange(trials, desc=desc)
    for trial in outer_it:
        random.shuffle(corpus)
        test_corpus = corpus[:chunk] if chunk else corpus
        if batching:
            it = zip_longest(*[iter(test_corpus)]*batching, fillvalue=None)
            for batch in tqdm(it, desc="Examples", unit_scale=True, leave=False):
                items = [x for x in batch if x is not None]
                with logging_redirect_tqdm():
                    start = time.time()
                    all_matches = func(items)
                    for matches in all_matches:
                        rv.append((trial, len(matches),
                                   (time.time() - start)/len(items)))
        else:
            for text, context in tqdm(test_corpus, desc="Examples", unit_scale=True, leave=False):
                with logging_redirect_tqdm():
                    start = time.time()
                    matches = func(text, context=context)
                    rv.append((trial, len(matches), time.time() - start))
    return rv


def iter_corpus():
    benchmarker = BioIDBenchmarker()
    for text, article in tqdm(benchmarker.processed_data[['text', 'don_article']].values):
        yield text, benchmarker._get_plaintext(article)


def build(trials: int, chunk: Optional[int] = None) -> pd.DataFrame:
    click.secho("Warming up python grounder")
    start = time.time()
    grounder.get_grounder()
    end = time.time() - start
    click.secho(f"Warmed up in {end:.2f} seconds")

    click.secho("Warming up local api grounder")
    ground_app_local_context("ER", context="Calcium is released from the ER.")

    click.secho("Warming up remote api grounder")
    ground_app_remote_context("ER", context="Calcium is released from the ER.")

    click.secho("Preparing BioID corpus")
    corpus = list(iter_corpus())

    rows = []
    for tag, uses_context, func, batching in FUNCTIONS:
        rv = run_trial(
            trials=trials,
            chunk=chunk,
            corpus=corpus,
            func=func,
            batching=batching,
            desc=f"{tag}{' with context' if uses_context else ''} trial",
        )
        rows.extend((tag, uses_context, *row) for row in rv)
    df = pd.DataFrame(rows, columns=["type", "context", "trial", "matches", "duration"])
    return df


@click.command()
@click.option("--trials", type=int, default=2, show_default=True)
@click.option(
    "--chunk",
    type=int,
    help="Subsample size from full corpus. Defaults to full corpus if not given.",
)
@verbose_option
@force_option
def main(trials: int, chunk: Optional[int], force: bool):
    if RESULTS_PATH.is_file() and not force:
        df = pd.read_csv(RESULTS_PATH, sep="\t")
    else:
        df = build(trials=trials, chunk=chunk)
        df.to_csv(RESULTS_PATH, sep="\t", index=False)

    # convert from seconds/response to responses/second
    df["duration"] = df["duration"].map(lambda x: 1 / x)

    _grouped = df[["type", "context", "duration"]].groupby(["type", "context"])
    agg_mean_df = _grouped.mean()
    agg_mean_df.rename(columns={"duration": "duration_mean"}, inplace=True)
    agg_std_df = _grouped.std()
    agg_std_df.rename(columns={"duration": "duration_std"}, inplace=True)
    agg_df = pd.merge(agg_mean_df, agg_std_df, left_index=True, right_index=True)
    agg_df = agg_df.round(1)
    agg_df.to_csv(RESULTS_AGG_PATH, sep="\t")
    agg_df.to_latex(
        RESULTS_AGG_TEX_PATH,
        label="tab:bioid-responsiveness-benchmark",
    )

    fig, ax = plt.subplots(figsize=(6, 3))
    sns.boxplot(data=df, y="duration", x="type", hue="context", ax=ax)
    ax.set_title("Gilda Responsiveness Benchmark on BioID")
    ax.set_yscale("log")
    ax.set_ylabel("Responses per Second")
    ax.set_xlabel("")
    fig.savefig(FIG_PATH)
    fig.savefig(FIG_PDF_PATH)


if __name__ == "__main__":
    main()



"""
Run the grounding app.

Run as module:
    `python -m gilda.app --host <host> --port <port> --terms <terms>`

Run with gunicorn:
    `gunicorn -w <worker count> -b <host>:<port> -t <timeout> gilda.app:gilda_app`

In case a non-standard set of terms is to be used, set the `GILDA_TERMS`
environment variable to the path to the terms file.
"""

import os
from .app import get_app, main


terms = os.environ.get('GILDA_TERMS')
gilda_app = get_app(terms=terms)



"""This script grounds entries in the MedMentions data set available at
https://github.com/chanzuckerberg/MedMentions. This can serve both as a
benchmark and as a bsis for creating mappings between the namespaces
Gilda grounds to and UMLS, which is used for MedMentions groundings.

.. code-block:: bibtex

    @article{Mohan2019,
        archivePrefix = {arXiv},
        arxivId = {1902.09476},
        author = {Mohan, Sunil and Li, Donghui},
        eprint = {1902.09476},
        month = {feb},
        title = {{MedMentions: A Large Biomedical Corpus Annotated with UMLS Concepts}},
        url = {http://arxiv.org/abs/1902.09476},
        year = {2019}
    }
"""

import csv
import json

import click
import gilda
import pyobo
import pystow
from more_click import verbose_option
from pubtator_loader import from_gz
from tqdm import tqdm
from tqdm.contrib.logging import logging_redirect_tqdm

URL = "https://github.com/chanzuckerberg/MedMentions/raw/master/full/data/corpus_pubtator.txt.gz"
MODULE = pystow.module("gilda", "medmentions")
CORPUS_PATH = MODULE.join(name="corpus.json")
MATCHING_PATH = pystow.join(
    "gilda", "medmentions", gilda.__version__, name="matching.tsv"
)

#: Subset of types from the Semantic Type Ontology used in annotation
TYPES = {
    "T116": "Amino Acid, Peptide, or Protein",
    "T123": "Biologically Active Substance",
    "T047": "Disease or Syndrome",
    "T101": "Patient or Disabled Group",
    "T079": "Temporal Concept",
    "T169": "Functional Concept",
    "T033": "Finding",
    "T081": "Quantitative Concept",
    "T063": "Molecular Biology Research Technique",
    "T052": "Activity",
    "T062": "Research Activity",
    "T032": "Organism Attribute",
    "T098": "Population Group",
    "T100": "Age Group",
    "T073": "Manufactured Object",
    "T093": "Health Care Related Organization",
    "T026": "Cell Component",
    "T028": "Gene or Genome",
    "T007": "Bacterium",
    "T045": "Genetic Function",
    "T046": "Pathologic Function",
    "T131": "Hazardous or Poisonous Substance",
    "T043": "Cell Function",
    "T025": "Cell",
    "T067": "Phenomenon or Process",
    "T069": "Environmental Effect of Humans",
    "T167": "Substance",
    "T037": "Injury or Poisoning",
    "T001": "Organism",
    "T031": "Body Substance",
    "T080": "Qualitative Concept",
    "T121": "Pharmacologic Substance",
    "T196": "Element, Ion, or Isotope",
    "T044": "Molecular Function",
    "T126": "Enzyme",
    "T077": "Conceptual Entity",
    "T082": "Spatial Concept",
    "T103": "Chemical",
    "T078": "Idea or Concept",
    "T039": "Physiologic Function",
    "T109": "Organic Chemical",
    "T122": "Biomedical or Dental Material",
    "T074": "Medical Device",
    "T058": "Health Care Activity",
    "T070": "Natural Phenomenon or Process",
    "T057": "Occupational Activity",
    "T023": "Body Part, Organ, or Organ Component",
    "T059": "Laboratory Procedure",
    "T042": "Organ or Tissue Function",
    "T060": "Diagnostic Procedure",
    "T184": "Sign or Symptom",
    "T024": "Tissue",
    "T029": "Body Location or Region",
    "T053": "Behavior",
    "T061": "Therapeutic or Preventive Procedure",
    "T083": "Geographic Area",
    "T055": "Individual Behavior",
    "T185": "Classification",
    "UnknownType": "UnknownType",
    "T054": "Social Behavior",
    "T170": "Intellectual Product",
    "T097": "Professional or Occupational Group",
    "T041": "Mental Process",
    "T095": "Self-help or Relief Organization",
    "T099": "Family Group",
    "T048": "Mental or Behavioral Dysfunction",
    "T071": "Entity",
    "T204": "Eukaryote",
    "T168": "Food",
    "T019": "Congenital Abnormality",
    "T201": "Clinical Attribute",
    "T040": "Organism Function",
    "T191": "Neoplastic Process",
    "T049": "Cell or Molecular Dysfunction",
    "T114": "Nucleic Acid, Nucleoside, or Nucleotide",
    "T002": "Plant",
    "T090": "Occupation or Discipline",
    "T096": "Group",
    "T075": "Research Device",
    "T034": "Laboratory or Test Result",
    "T197": "Inorganic Chemical",
    "T195": "Antibiotic",
    "T091": "Biomedical Occupation or Discipline",
    "T008": "Animal",
    "T102": "Group Attribute",
    "T012": "Bird",
    "T011": "Amphibian",
    "T200": "Clinical Drug",
    "T125": "Hormone",
    "T129": "Immunologic Factor",
    "T127": "Vitamin",
    "T056": "Daily or Recreational Activity",
    "T087": "Amino Acid Sequence",
    "T086": "Nucleotide Sequence",
    "T005": "Virus",
    "T065": "Educational Activity",
    "T192": "Receptor",
    "T022": "Body System",
    "T015": "Mammal",
    "T038": "Biologic Function",
    "T130": "Indicator, Reagent, or Diagnostic Aid",
    "T030": "Body Space or Junction",
    "T017": "Anatomical Structure",
    "T068": "Human-caused Phenomenon or Process",
    "T016": "Human",
    "T050": "Experimental Model of Disease",
    "T089": "Regulation or Law",
    "T085": "Molecular Sequence",
    "T104": "Chemical Viewed Structurally",
    "T064": "Governmental or Regulatory Activity",
    "T190": "Anatomical Abnormality",
    "T092": "Organization",
    "T051": "Event",
    "T004": "Fungus",
    "T171": "Language",
    "T014": "Reptile",
    "T120": "Chemical Viewed Functionally",
    "T020": "Acquired Abnormality",
    "T013": "Fish",
    "T194": "Archaeon",
    "T072": "Physical Object",
    "T018": "Embryonic Structure",
    "T066": "Machine Activity",
    "T010": "Vertebrate",
    "T094": "Professional Society",
    "T021": "Fully Formed Anatomical Structure",
    "T203": "Drug Delivery Device",
}


def get_corpus():
    """Get the MedMentions corpus.

    :return: A list of dictionaries of the following form:

    .. code-block::

        [
          {
            "id": 25763772,
            "title_text": "DCTN4 as a modifier of chronic Pseudomonas aeruginosa infection in cystic fibrosis",
            "abstract_text": "...",
            "entities": [
              {
                "document_id": 25763772,
                "start_index": 0,
                "end_index": 5,
                "text_segment": "DCTN4",
                "semantic_type_id": "T116,T123",
                "entity_id": "C4308010"
              },
              ...
            ]
          },
          ...
        ]
    """
    if not CORPUS_PATH.is_file():
        path = MODULE.ensure(url=URL)
        corpus = from_gz(path)
        with CORPUS_PATH.open("w") as file:
            json.dump(corpus, file, indent=2, default=lambda o: o.__dict__)

    # Right now I'd rather not engage with the strange object structure, so
    # serializing and deserializing gets us JSON we can work with.
    with CORPUS_PATH.open() as file:
        return json.load(file)


def iterate_corpus():
    corpus = get_corpus()
    click.echo(f"MedMentions has {len(corpus)} entries")
    for document in tqdm(corpus, unit="document", desc="MedMentions"):
        document_id = document["id"]
        abstract = document["abstract_text"]
        for entity_idx, entity in enumerate(document["entities"]):
            umls_id = entity["entity_id"]
            text = entity["text_segment"]
            start, end = entity["start_index"], entity["end_index"]
            types = set(entity["semantic_type_id"].split(","))
            yield document_id, abstract, umls_id, text, start, end, types


HEADER = [
    "document",
    "start_idx",
    "end_idx",
    "text",
    "umls_id",
    "umls_name",
    "gilda_prefix",
    "gilda_identifier",
    "gilda_name",
    "gilda_score",
]


@click.command()
@verbose_option
def main():
    corpus = get_corpus()
    click.echo(f"There are {len(corpus)} entries")
    rows = []
    for document_id, abstract, umls_id, text, start, end, types in iterate_corpus():
        with logging_redirect_tqdm():
            matches = gilda.ground(text, context=abstract)
        rows.extend(
            (
                document_id,
                start,
                end,
                text,
                umls_id,
                pyobo.get_name("umls", umls_id),
                match.term.db,
                match.term.id,
                match.term.entry_name,
                match.score,
            )
            for match in matches
        )
    with MATCHING_PATH.open("w") as file:
        writer = csv.writer(file, delimiter="\t")
        writer.writerow(HEADER)
        writer.writerows(rows)


if __name__ == "__main__":
    main()



"""This is a script that can be run to generate a new grounding_terms.tsv file.
It uses several resource files and database clients from INDRA and requires it
to be available locally."""

import re
import os
import csv
import json
import logging
import requests
import indra
from indra.databases import hgnc_client, uniprot_client, chebi_client, \
    go_client, mesh_client, doid_client
from indra.statements.resources import amino_acids
from gilda.term import Term, dump_terms, filter_out_duplicates
from gilda.process import normalize
from gilda.resources import resource_dir, popular_organisms


indra_module_path = indra.__path__[0]
indra_resources = os.path.join(indra_module_path, 'resources')

logger = logging.getLogger('gilda.generate_terms')


def read_csv(fname, header=False, delimiter='\t', quotechar='"'):
    with open(fname, 'r') as fh:
        reader = csv.reader(fh, delimiter=delimiter, quotechar=quotechar)
        if header:
            header_names = next(reader)
            for row in reader:
                yield {h: r for h, r in zip(header_names, row)}
        else:
            for row in reader:
                yield row


def generate_hgnc_terms():
    fname = os.path.join(resource_dir, 'hgnc_entries.tsv')

    if not os.path.exists(fname):
        # Select relevant columns and parameters
        cols = [
            'gd_hgnc_id', 'gd_app_sym', 'gd_app_name', 'gd_status',
            'gd_aliases', 'gd_prev_sym', 'gd_name_aliases'
        ]

        statuses = ['Approved', 'Entry%20Withdrawn']
        params = {
                'hgnc_dbtag': 'on',
                'order_by': 'gd_app_sym_sort',
                'format': 'text',
                'submit': 'submit'
                }

        # Construct a download URL from the above parameters
        url = 'https://www.genenames.org/cgi-bin/download/custom?'
        url += '&'.join(['col=%s' % c for c in cols]) + '&'
        url += '&'.join(['status=%s' % s for s in statuses]) + '&'
        url += '&'.join(['%s=%s' % (k, v) for k, v in params.items()])

        # Download the file
        logger.info('Downloading HGNC resource file')
        res = requests.get(url)
        with open(fname, 'w') as fh:
            fh.write(res.text)

    logger.info('Loading %s' % fname)
    all_term_args = {}
    rows = [r for r in read_csv(fname, header=True, delimiter='\t',
                                quotechar=None)]
    id_name_map = {r['HGNC ID'].split(':')[1]: r['Approved symbol']
                   for r in rows}
    organism = '9606'  # human
    for row in rows:
        db, id = row['HGNC ID'].split(':')
        name = row['Approved symbol']
        # Special handling for rows representing withdrawn symbols
        if row['Status'] == 'Symbol Withdrawn':
            m = re.match(r'symbol withdrawn, see \[HGNC:(?: ?)(\d+)\]',
                         row['Approved name'])
            new_id = m.groups()[0]
            new_name = id_name_map[new_id]
            term_args = (normalize(name), name, db, new_id,
                         new_name, 'former_name', 'hgnc', organism)
            all_term_args[term_args] = None
            # NOTE: consider adding withdrawn synonyms e.g.,
            # symbol withdrawn, see pex1     symbol withdrawn, see PEX1
            # HGNC    13197   ZWS1~withdrawn  synonym
            continue
        # Handle regular entry official names
        else:
            term_args = (normalize(name), name, db, id, name, 'name', 'hgnc',
                         organism)
            all_term_args[term_args] = None
            if row['Approved name']:
                app_name = row['Approved name']
                term_args = (normalize(app_name), app_name, db, id, name,
                             'name', 'hgnc', organism)
                all_term_args[term_args] = None

        # Handle regular entry synonyms
        synonyms = []
        if row['Alias symbols']:
            synonyms += row['Alias symbols'].split(', ')
        for synonym in synonyms:
            term_args = (normalize(synonym), synonym, db, id, name, 'synonym',
                         'hgnc', organism)
            all_term_args[term_args] = None

        # Handle regular entry previous symbols
        if row['Previous symbols']:
            prev_symbols = row['Previous symbols'].split(', ')
            for prev_symbol in prev_symbols:
                term_args = (normalize(prev_symbol), prev_symbol, db, id, name,
                             'former_name', 'hgnc', organism)
                all_term_args[term_args] = None

        if row['Alias names']:
            for alias_name in extract_hgnc_alias_names(row['Alias names']):
                alias_name = alias_name.strip()
                # There are double quotes and sometimes spurious extra spaces
                term_args = (normalize(alias_name), alias_name, db, id, name,
                             'synonym', 'synonym', organism)
                all_term_args[term_args] = None

    terms = [Term(*args) for args in all_term_args.keys()]
    logger.info('Loaded %d terms' % len(terms))
    return terms


def extract_hgnc_alias_names(alias_str):
    # The string is a comma-separated list of aliases each within
    # double quotes, and commas can appear within double quotes as well.
    if re.match(r'^"([^"]+)"$', alias_str):
        names = [alias_str.strip('"').strip()]
    else:
        names = [s.strip() for s in
                 next(csv.reader([alias_str], skipinitialspace=True))]
    return names


def generate_chebi_terms():
    # We can get standard names directly from the OBO
    terms = _generate_obo_terms('chebi', ignore_mappings=True,
                                map_to_ns={})

    # Now we add synonyms
    # NOTE: this file is not in version control. The file is available
    # at ftp://ftp.ebi.ac.uk/pub/databases/chebi/Flat_file_
    # tab_delimited/names_3star.tsv.gz, it needs to be decompressed
    # into the INDRA resources folder.
    fname = os.path.join(indra_resources, 'names_3star.tsv')
    if not os.path.exists(fname):
        import pandas as pd
        chebi_url = 'ftp://ftp.ebi.ac.uk/pub/databases/chebi/' \
                    'Flat_file_tab_delimited/names_3star.tsv.gz'
        logger.info('Loading %s into memory. You can download and decompress'
                    ' it in the indra/resources folder for faster access.'
                    % chebi_url)
        df = pd.read_csv(chebi_url, sep='\t')
        rows = (row for _, row in df.iterrows())
    else:
        rows = read_csv(fname, header=True, delimiter='\t')

    added = set()
    for row in rows:
        chebi_id = chebi_client.get_primary_id(str(row['COMPOUND_ID']))
        if not chebi_id:
            logger.info('Could not get valid CHEBI ID for %s' %
                        row['COMPOUND_ID'])
            continue
        db = 'CHEBI'
        name = str(row['NAME'])
        chebi_name = \
            chebi_client.get_chebi_name_from_id(chebi_id, offline=True)
        if chebi_name is None:
            logger.info('Could not get valid name for %s' % chebi_id)
            continue
        # We skip entries of the form Glu-Lys with synonyms like EK since
        # there are highly ambiguous with other acronyms, and are unlikely
        # to be used in practice.
        if is_aa_sequence(chebi_name) and re.match(r'(^[A-Z-]+$)', name):
            continue

        term_args = (normalize(name), name, db, chebi_id, chebi_name, 'synonym',
                     'chebi')
        if term_args in added:
            continue
        else:
            term = Term(*term_args)
            terms.append(term)
            added.add(term_args)
    logger.info('Loaded %d terms' % len(terms))
    return terms


def is_aa_sequence(txt):
    """Return True if the given text is a sequence of amino acids like Tyr-Glu.
    """
    return ('-' in txt) and (all(part in aa_abbrevs
                                 for part in txt.split('-')))


aa_abbrevs = {aa['short_name'].capitalize() for aa in amino_acids.values()}


def generate_mesh_terms(ignore_mappings=False):
    mesh_name_files = ['mesh_id_label_mappings.tsv',
                       'mesh_supp_id_label_mappings.tsv']
    terms = []
    for fname in mesh_name_files:
        logger.info('Loading %s' % fname)
        mesh_names_file = os.path.join(indra_resources, fname)
        for row in read_csv(mesh_names_file, header=False, delimiter='\t'):
            db_id = row[0]
            text_name = row[1]
            mapping = mesh_mappings.get(db_id)
            if not ignore_mappings and mapping and mapping[0] \
                    not in {'EFO', 'HP', 'DOID'}:
                db, db_id, name = mapping
                status = 'synonym'
            else:
                db = 'MESH'
                status = 'name'
                name = text_name
            term = Term(normalize(text_name), text_name, db, db_id, name,
                        status, 'mesh',
                        source_db='MESH' if db != 'MESH' else None,
                        source_id=row[0] if db != 'MESH' else None)
            terms.append(term)
            synonyms = row[2]
            if row[2]:
                synonyms = synonyms.split('|')
                for synonym in synonyms:
                    term = Term(normalize(synonym), synonym, db, db_id, name,
                                'synonym', 'mesh',
                                source_db='MESH' if db != 'MESH' else None,
                                source_id=row[0] if db != 'MESH' else None)
                    terms.append(term)
        logger.info('Loaded %d terms' % len(terms))
    return terms


def generate_go_terms():
    fname = os.path.join(indra_resources, 'go.json')
    logger.info('Loading %s' % fname)
    with open(fname, 'r') as fh:
        entries = json.load(fh)
    terms = []
    for entry in entries:
        go_id = entry['id']
        name = entry['name']
        # First handle the name term
        term = Term(normalize(name), name, 'GO', go_id, name, 'name', 'go')
        terms.append(term)
        # Next look at synonyms, sometimes those match the name so we
        # deduplicate
        for synonym in set(entry.get('synonyms', [])) - {name}:
            # GO includes around 40k synonyms for terms that represent
            # activity out of which around 5k are actually
            # synonyms for entities. One example is "EGFR" as a synonym
            # for "epidermal growth factor-activated receptor activity".
            # We skip these according to the following logic.
            if 'activity' in name and 'activity' not in synonym:
                continue
            term = Term(normalize(synonym), synonym, 'GO', go_id, name,
                        'synonym', 'go')
            terms.append(term)
    logger.info('Loaded %d terms' % len(terms))
    return terms


def generate_famplex_terms(ignore_mappings=False):
    fname = os.path.join(indra_resources, 'famplex', 'grounding_map.csv')
    logger.info('Loading %s' % fname)
    terms = []
    for row in read_csv(fname, delimiter=','):
        txt = row[0]
        norm_txt = normalize(txt)
        groundings = {k: v for k, v in zip(row[1::2], row[2::2]) if (k and v)}
        if 'FPLX' in groundings:
            id = groundings['FPLX']
            term = Term(norm_txt, txt, 'FPLX', id, id, 'curated', 'famplex')
        elif 'HGNC' in groundings:
            id = groundings['HGNC']
            term = Term(norm_txt, txt, 'HGNC', hgnc_client.get_hgnc_id(id), id,
                        'curated', 'famplex', '9606')
        elif 'UP' in groundings:
            db = 'UP'
            id = groundings['UP']
            name = id
            organism = None
            if uniprot_client.is_human(id):
                organism = '9606'
                hgnc_id = uniprot_client.get_hgnc_id(id)
                if hgnc_id:
                    name = hgnc_client.get_hgnc_name(hgnc_id)
                    if hgnc_id:
                        db = 'HGNC'
                        id = hgnc_id
                else:
                    logger.warning('No gene name for %s' % id)
            # FIXME: we should figure out what organism the given protein
            # comes from and then add that organism info, otherwise
            # these groundings will be asserted even if the organism
            # doesn't match
            term = Term(norm_txt, txt, db, id, name, 'curated', 'famplex',
                        organism)
        elif 'CHEBI' in groundings:
            id = groundings['CHEBI']
            name = chebi_client.get_chebi_name_from_id(id[6:])
            term = Term(norm_txt, txt, 'CHEBI', id, name, 'curated',
                        'famplex')
        elif 'GO' in groundings:
            id = groundings['GO']
            term = Term(norm_txt, txt, 'GO', id,
                        go_client.get_go_label(id), 'curated', 'famplex')
        elif 'MESH' in groundings:
            id = groundings['MESH']
            mesh_mapping = mesh_mappings.get(id)
            db, db_id, name = mesh_mapping if (mesh_mapping
                                               and not ignore_mappings) else \
                ('MESH', id, mesh_client.get_mesh_name(id))
            term = Term(norm_txt, txt, db, db_id, name, 'curated', 'famplex',
                        None, 'MESH', groundings['MESH'])
        else:
            # TODO: handle HMDB, PUBCHEM, CHEMBL
            continue
        terms.append(term)
    return terms


def generate_uniprot_terms(download=False, organisms=None):
    if not organisms:
        organisms = popular_organisms
    path = os.path.join(resource_dir, 'up_synonyms.tsv')
    if not os.path.exists(path) or download:
        # Columns according to the new API, comments for old API
        columns = [
            'accession',     # id
            'gene_primary',  # genes(PREFERRED)
            'gene_synonym',  # genes(ALTERNATIVE)'
            'protein_name',  # protein names
            'organism_id',   # organism-id
        ]
        org_filter_str = '+OR+'.join(f'(taxonomy_id:{org})' for org in organisms)
        query = f'reviewed:true+AND+({org_filter_str})'
        url = (f'https://rest.uniprot.org/uniprotkb/stream?'
               f'format=tsv&'
               f'query={query}&'
               f'compressed=false&'
               f'fields={",".join(columns)}')
        logger.info('Downloading UniProt resource file')
        res = requests.get(url)
        with open(path, 'w') as fh:
            fh.write(res.text)
    terms = []
    logger.info('Loading %s' % path)
    for row in read_csv(path, delimiter='\t', header=True):
        terms += get_terms_from_uniprot_row(row)

    return terms


def get_terms_from_uniprot_row(row):
    terms = []
    up_id = row['Entry']
    organism = row['Organism (ID)']

    # As of 3/2/2022 there is an error in UniProt data that we need to manually
    # patch here
    if up_id == 'Q2QKR2':
        row['Protein names'] = row['Protein names'][:-1]
    protein_names = parse_uniprot_synonyms(row['Protein names'])

    # These two lists are aligned and each separated by "; " if there
    # are multiple genes. If there are no genes listed, we simply have
    # an empty string here. If there are multiple genes but one doesn't
    # have a name given, a corresponding "; " placeholder is still there.
    # Consequently, we encounter cases like
    # P34539»·; »·; »·
    # where there are two genes but neither of them have names or
    # synonyms listed.
    primary_gene_names = row['Gene Names (primary)'].split('; ')
    gene_synonyms = row['Gene Names (synonym)'].split('; ')

    multi_gene = len(primary_gene_names) > 1

    # We generally use the gene name as the standard name
    # except when there are multiple gene names or the gene name is missing
    if not primary_gene_names or multi_gene:
        standard_name = protein_names[0]
    else:
        standard_name = primary_gene_names[0]
    # We skip a small number of non-critical entries that don't have
    # standard names
    if not standard_name:
        return []

    # By default, we use the UniProt namespace and ID
    ns = 'UP'
    id = up_id
    # For human genes, we resolve redundancies by mapping to HGNC
    hgnc_id = uniprot_client.get_hgnc_id(up_id)
    # We only map to HGNC if there is a single gene corresponding to
    # this protein. Otherwise, we keep the UniProt ID.
    if hgnc_id and not multi_gene:
        ns = 'HGNC'
        id = hgnc_id
        standard_name = hgnc_client.get_hgnc_name(hgnc_id)

    # We add all of the protein names as synonyms
    for name in protein_names:
        # Skip names that are EC codes
        if name.startswith('EC '):
            continue
        if name == standard_name:
            continue
        term = Term(normalize(name), name, ns, id,
                    standard_name, 'synonym', 'uniprot',
                    organism, None if ns == 'UP' else 'UP',
                    None if ns == 'UP' else up_id)
        terms.append(term)

    # We add the standard name (usually the gene name)
    term = Term(normalize(standard_name), standard_name,
                ns, id, standard_name, 'name', 'uniprot',
                organism, None if ns == 'UP' else 'UP',
                None if ns == 'UP' else up_id)
    terms.append(term)

    # If we have gene synonyms we include them according to the following logic.
    # For human genes we do not include synonyms if there are multiple genes
    # corresponding to this protein. For non-human genes, since we don't
    # have a separate gene resource, we do include gene names and synonyms
    # even if there are multiple genes corresponding to the protein.
    if not ((organism == '9606') and multi_gene):
        for gene_name, gene_synonyms_str in zip(primary_gene_names,
                                                gene_synonyms):
            all_synonyms = gene_synonyms_str.split(' ')
            # We can skip a gene name if we used it as standard name
            if gene_name and (gene_name != standard_name):
                all_synonyms.append(gene_name)
            for synonym in all_synonyms:
                if not synonym:
                    continue
                term = Term(normalize(synonym), synonym,
                            ns, id, standard_name, 'synonym', 'uniprot',
                            organism, None if ns == 'UP' else 'UP',
                            None if ns == 'UP' else up_id)
                terms.append(term)
    return terms


ec_code_pattern = re.compile(r'EC [\d\.-]+')


def parse_uniprot_synonyms(synonyms_str):
    synonyms_str = re.sub(r'\[Includes: ([^]])+\]',
                          '', synonyms_str).strip()
    synonyms_str = re.sub(r'\[Cleaved into: ([^]])+\]( \(Fragments\))?',
                          '', synonyms_str).strip()

    def find_block_from_right(s):
        parentheses_depth = 0
        assert s.endswith(')')
        s = s[:-1]
        block = ''
        for c in s[::-1]:
            if c == ')':
                parentheses_depth += 1
            elif c == '(':
                if parentheses_depth > 0:
                    parentheses_depth -= 1
                else:
                    return block
            block = c + block
        return block

    syns = []
    while True:
        # If the string is empty at this point, we return with the
        # synonyms so far
        if not synonyms_str:
            return syns
        # If the string doesn't end with a parenthesis, that means it's
        # the first synonym which isn't in parentheses, and so we prepend
        # it to the list of synonyms and return
        if not synonyms_str.endswith(')'):
            return [synonyms_str] + syns

        syn = find_block_from_right(synonyms_str)
        # This is where we remove the synonym we just processed plus
        # the space and opening/closing parens (3 characters) to get to the
        # next synonym
        to_remove = len(syn) + 3
        # One corner case is when there happens to be a parenthesis in the
        # first synonym in the list like "X(0) (...) (...)"
        # In this case the entire synonyms_str is to be used and we can
        # prepend and return
        if synonyms_str[-to_remove] != ' ':
            return [synonyms_str] + syns

        # EC codes are not valid synonyms bot otherwise we prepend
        # the synonym to the list of synonyms
        if not ec_code_pattern.match(syn):
            syns = [syn] + syns
        # We now remove the processed suffix
        synonyms_str = synonyms_str[:-to_remove]


def generate_adeft_terms():
    from adeft import available_shortforms
    from adeft.disambiguate import load_disambiguator
    from indra.ontology.standardize import get_standard_name
    all_term_args = set()
    add_prefix = ['BTO', 'HP', 'DOID']
    remove_prefix = ['EFO', 'NCIT', 'OMIT']
    for shortform in available_shortforms:
        da = load_disambiguator(shortform)
        for grounding, name in da.names.items():
            if grounding == 'ungrounded' or ':' not in grounding:
                continue
            db_ns, db_id = grounding.split(':', maxsplit=1)
            if db_ns in remove_prefix:
                if db_id.startswith(db_ns + ':'):
                    db_id = db_id[len(db_ns) + 1:]
            if db_ns in add_prefix:
                if not db_id.startswith(db_ns + ':'):
                    db_id = db_ns + ':' + db_id
            if db_id == 'PF00112)':
                db_id = 'PF00112'
            # Here we do a name standardization via INDRA just in case
            # there is a discrepancy
            indra_standard_name = get_standard_name({db_ns: db_id})
            if indra_standard_name:
                name = indra_standard_name
            term_args = (normalize(shortform), shortform, db_ns, db_id,
                         name, 'synonym', 'adeft')
            all_term_args.add(term_args)
    terms = [Term(*term_args) for term_args in sorted(list(all_term_args),
                                                      key=lambda x: x[0])]
    return terms


def generate_doid_terms(ignore_mappings=False):
    return _generate_obo_terms('doid', ignore_mappings)


def generate_efo_terms(ignore_mappings=False):
    terms = _generate_obo_terms('efo', ignore_mappings)
    # We remove BFO terms since they are too generic to be useful
    terms = [t for t in terms if not t.id.startswith('BFO:')]
    return terms


def generate_hp_terms(ignore_mappings=False):
    return _generate_obo_terms('hp', ignore_mappings)


def terms_from_obo_json_entry(entry, prefix, ignore_mappings=False,
                              map_to_ns=None):
    if map_to_ns is None:
        map_to_ns = {'MESH', 'DOID'}
    terms = []
    db, db_id, name = prefix.upper(), entry['id'], entry['name']
    entry_name = name
    # We first need to decide if we prioritize another name space
    xref_dict = {xr['namespace']: xr['id'] for xr in entry.get('xrefs', [])}
    # Handle MeSH mappings first
    auto_mesh_mapping = mesh_mappings_reverse.get((db, db_id))
    if auto_mesh_mapping and not ignore_mappings:
        db, db_id, entry_name = ('MESH', auto_mesh_mapping[0],
                                 auto_mesh_mapping[1])
    elif 'MESH' in map_to_ns and ('MESH' in xref_dict or 'MSH' in xref_dict):
        mesh_id = xref_dict.get('MESH') or xref_dict.get('MSH')
        # Since we currently only include regular MeSH terms (which start
        # with D), we only need to do the mapping if that's the case.
        # We don't map any supplementary terms that start with C.
        if mesh_id.startswith('D'):
            mesh_name = mesh_client.get_mesh_name(mesh_id)
            if mesh_name:
                # Here we need to check if we further map the MeSH ID to
                # another namespace
                mesh_mapping = mesh_mappings.get(mesh_id)
                db, db_id, entry_name = mesh_mapping if \
                    (mesh_mapping and (mesh_mapping[0]
                                       not in {'EFO', 'HP', 'DOID'})) \
                    else ('MESH', mesh_id, mesh_name)
    # Next we look at mappings to DOID
    # TODO: are we sure that the DOIDs that we get here (from e.g., EFO)
    # cannot be mapped further to MeSH per the DOID resource file?
    elif 'DOID' in map_to_ns and 'DOID' in xref_dict:
        doid = xref_dict['DOID']
        if not doid.startswith('DOID:'):
            doid = 'DOID:' + doid
        doid_prim_id = doid_client.get_doid_id_from_doid_alt_id(doid)
        if doid_prim_id:
            doid = doid_prim_id
        doid_name = doid_client.get_doid_name_from_doid_id(doid)
        # If we don't get a name here, it's likely because an entry is
        # obsolete so we don't do the mapping
        if doid_name:
            db, db_id, entry_name = 'DOID', doid, doid_name

    # Add a term for the name first
    name_term = Term(
        norm_text=normalize(name),
        text=name,
        db=db,
        id=db_id,
        entry_name=entry_name,
        status='name',
        source=prefix,
        source_db=prefix.upper() if db != prefix.upper() else None,
        source_id=entry['id'] if db != prefix.upper() else None,
    )
    terms.append(name_term)

    # Then add all the synonyms
    for synonym in set(entry.get('synonyms', [])):
        # Some synonyms are tagged as ambiguous, we remove these
        if 'ambiguous' in synonym.lower():
            continue
        # Some synonyms contain a "formerly" clause, we remove these
        match = re.match(r'(.+) \(formerly', synonym)
        if match:
            synonym = match.groups()[0]
        # Some synonyms contain additional annotations
        # e.g. Hyperplasia of facial adipose tissue" NARROW
        # [ORCID:0000-0001-5889-4463]
        # If this is the case, we strip these off
        match = re.match(r'([^"]+)', synonym)
        if match:
            synonym = match.groups()[0]

        synonym_term = Term(
            norm_text=normalize(synonym),
            text=synonym,
            db=db,
            id=db_id,
            entry_name=entry_name,
            status='synonym',
            source=prefix,
            source_db=prefix.upper() if db != prefix.upper() else None,
            source_id=entry['id'] if db != prefix.upper() else None,
        )
        terms.append(synonym_term)
    return terms


def _generate_obo_terms(prefix, ignore_mappings=False, map_to_ns=None):
    filename = os.path.join(indra_resources, '%s.json' % prefix)
    logger.info('Loading %s', filename)
    with open(filename) as file:
        entries = json.load(file)
    terms = []
    for entry in entries:
        terms += terms_from_obo_json_entry(entry, prefix=prefix,
                                           ignore_mappings=ignore_mappings,
                                           map_to_ns=map_to_ns)
    logger.info('Loaded %d terms from %s', len(terms), prefix)
    return terms


def _make_mesh_mappings():
    # Load MeSH ID/label mappings
    from gilda.resources import MESH_MAPPINGS_PATH
    mesh_mappings = {}
    mesh_mappings_reverse = {}
    for row in read_csv(MESH_MAPPINGS_PATH, delimiter='\t'):
        # We can skip row[2] which is the MeSH standard name for the entry
        mesh_mappings[row[1]] = row[3:]
        mesh_mappings_reverse[(row[3], row[4])] = [row[1], row[2]]
    return mesh_mappings, mesh_mappings_reverse


mesh_mappings, mesh_mappings_reverse = _make_mesh_mappings()


def terms_from_obo_url(url, prefix, ignore_mappings=False, map_to_ns=None):
    """Return terms extracted directly from an OBO given as a URL."""
    import obonet
    from indra.databases.obo_client import OboClient
    g = obonet.read_obo(url)
    entries = OboClient.entries_from_graph(g, prefix=prefix)
    terms = []
    for entry in entries:
        terms += terms_from_obo_json_entry(entry, prefix=prefix,
                                           ignore_mappings=ignore_mappings,
                                           map_to_ns=map_to_ns)
    return terms


def get_all_terms():
    terms = []

    generated_term_groups = [
        generate_uniprot_terms(),
        generate_famplex_terms(),
        generate_hgnc_terms(),
        generate_chebi_terms(),
        generate_go_terms(),
        generate_mesh_terms(),
        generate_adeft_terms(),
        generate_doid_terms(),
        generate_hp_terms(),
        generate_efo_terms(),
    ]
    for generated_terms in generated_term_groups:
        terms += generated_terms
    terms = filter_out_duplicates(terms)
    return terms


def main():
    from gilda.resources import GROUNDING_TERMS_PATH as fname
    terms = get_all_terms()
    dump_terms(terms, fname)


if __name__ == '__main__':
    main()



from textwrap import dedent

import gilda
from gilda.ner import get_brat


def test_annotate():
    full_text = \
        "The protein BRAF is a kinase.\nBRAF is a gene.\nBRAF is a protein."

    annotations = gilda.annotate(full_text)
    assert isinstance(annotations, list)

    # Check that we get 4 annotations
    assert len(annotations) == 4

    # Check that the annotations are for the expected words
    assert tuple(a.text for a in annotations) == (
        'BRAF', 'kinase', 'BRAF', 'BRAF')

    # Check that the spans are correct
    expected_spans = ((12, 16), (22, 28), (30, 34), (46, 50))
    actual_spans = tuple((a.start, a.end) for a in annotations)
    assert actual_spans == expected_spans

    # Check that the curies are correct
    expected_curies = ("hgnc:1097", "mesh:D010770", "hgnc:1097", "hgnc:1097")
    actual_curies = tuple(a.matches[0].term.get_curie() for a in annotations)
    assert actual_curies == expected_curies


def test_get_brat():
    full_text = \
        "The protein BRAF is a kinase.\nBRAF is a gene.\nBRAF is a protein."

    brat_str = get_brat(gilda.annotate(full_text))

    assert isinstance(brat_str, str)
    match_str = dedent("""
        T1\tEntity 12 16\tBRAF
        #1\tAnnotatorNotes T1\thgnc:1097
        T2\tEntity 22 28\tkinase
        #2\tAnnotatorNotes T2\tmesh:D010770
        T3\tEntity 30 34\tBRAF
        #3\tAnnotatorNotes T3\thgnc:1097
        T4\tEntity 46 50\tBRAF
        #4\tAnnotatorNotes T4\thgnc:1097
        """).lstrip()
    assert brat_str == match_str


def test_get_all():
    full_text = "This is about ER."
    results = gilda.annotate(full_text)
    assert len(results) == 1
    curies = set()
    for annotation in results:
        for scored_match in annotation.matches:
            curies.add(scored_match.term.get_curie())
    assert "hgnc:3467" in curies  # ESR1
    assert "fplx:ESR" in curies
    assert "go:0005783" in curies  # endoplasmic reticulum


def test_context_test():
    text = "This is about ER."
    context_text = "Estrogen receptor (ER) is a protein family."
    results = gilda.annotate(text, context_text=context_text)
    assert len(results) == 1
    assert results[0].matches[0].term.get_curie() == "fplx:ESR"
    assert results[0].text == "ER"
    assert (results[0].start, results[0].end) == (14, 16)

    context_text = "Calcium is released from the ER."
    results = gilda.annotate(text, context_text=context_text)
    assert len(results) == 1
    assert results[0].matches[0].term.get_curie() == "go:0005783"
    assert results[0].text == "ER"
    assert (results[0].start, results[0].end) == (14, 16)


def test_punctuation_comma_in_entity():
    # A named entity with an actual comma in its name
    res = gilda.annotate('access, internet')
    assert len(res) == 1
    # Make sure we capture the text span exactly despite
    # tokenization
    assert res[0].text == 'access, internet'
    assert res[0].start == 0
    assert res[0].end == 16
    assert res[0].matches[0].term.db == 'MESH'
    assert res[0].matches[0].term.id == 'D000077230'


def test_punctuation_outside_entities():
    res = gilda.annotate('EGF binds EGFR, which is a receptor.')
    assert len(res) == 3

    assert [ann.text for ann in res] == ['EGF', 'EGFR', 'receptor']

    res = gilda.annotate('EGF binds EGFR: a receptor.')
    assert len(res) == 3

    assert [ann.text for ann in res] == ['EGF', 'EGFR', 'receptor']


def test_synonym_corner_case():
    # Context: EFO contains the synonyms "transthyretin " with a trailing
    # space, this triggers an indexing error corner case when followed by -
    res = gilda.annotate('transthyretin -')
    assert res


def test_cell_death():
    anns = gilda.annotate("cell death")
    assert len(anns) == 1
    ann = anns[0]
    # Make sure that we match the entire span, including
    # the "cell" part that is in itself a stopword
    assert ann.text == "cell death", ann.text


def test_a_gene():
    # This tests that we don't ground spans that start with
    # a core stop word like a, the, etc.
    anns = gilda.annotate("a gene")
    assert not anns


def test_ui_example_annotation():
    text = (
        "Small G proteins are an extensive family of proteins that bind and "
        "hydrolyze GTP. They are ubiquitous inside cells, regulating a wide range "
        "of cellular processes. Recently, many studies have examined the role of "
        "small G proteins, particularly the Ras family of G proteins, in memory "
        "formation. Once thought to be primarily involved in the transduction of a "
        "variety of extracellular signals during development, it is now clear that "
        "Ras family proteins also play critical roles in molecular processing "
        "underlying neuronal and behavioral plasticity. We here review a number of "
        "recent studies that explore how the signaling of Ras family proteins "
        "contributes to memory formation. Understanding these signaling processes "
        "is of fundamental importance both from a basic scientific perspective, "
        "with the goal of providing mechanistic insights into a critical aspect of "
        "cognitive behavior, and from a clinical perspective, with the goal of "
        "providing effective therapies for a range of disorders involving cognitive "
        "impairments."
    )
    annotations = gilda.annotate(text)
    expected_annotations = [
        # Start, End, Text, Grounding
        (0, 16, "Small G proteins", "hgnc:9802"),
        (77, 80, "GTP", "chebi:15996"),
        (143, 161, "cellular processes", "go:0009987"),
        (212, 228, "small G proteins", "hgnc:9802"),
        (247, 250, "Ras", "fplx:RAS"),
        (261, 271, "G proteins", "fplx:G_protein"),
        (276, 282, "memory", "go:0007613"),
        (339, 351, "transduction", "go:0009293"),
        (368, 381, "extracellular", "go:0005576"),
        (431, 434, "Ras", "fplx:RAS"),
        (610, 619, "signaling", "go:0023052"),
        (623, 626, "Ras", "fplx:RAS"),
        (658, 664, "memory", "go:0007613"),
        (676, 689, "Understanding", "mesh:D032882"),
        (696, 715, "signaling processes", "go:0023052"),
        (871, 879, "behavior", "go:0007610"),
        (951, 960, "therapies", "mesh:D013812"),
        (976, 985, "disorders", "mesh:D004194"),
        (996, 1017, "cognitive impairments", "mesh:D060825"),
    ]
    assert len(annotations) == len(expected_annotations)
    for ann, expected in zip(annotations, expected_annotations):
        exp_start, exp_end, exp_text, exp_grounding = expected
        assert ann.start == exp_start
        assert ann.end == exp_end
        assert ann.text == exp_text
        assert ann.matches[0].term.get_curie() == exp_grounding



# -*- coding: utf-8 -*-

"""This script measures the responsiveness (i.e., speed) of Gilda on the
MedMentions corpus in three settings: when used as a python package,
a local web service or through the remote public web service.
"""

import pathlib
import random
import time
from textwrap import dedent
from typing import Optional

import click
import gilda
import matplotlib.pyplot as plt
import pandas as pd
import requests
import seaborn as sns
from gilda.api import grounder
from more_click import force_option, verbose_option
from tqdm import tqdm, trange
from tqdm.contrib.logging import logging_redirect_tqdm

from medmentions import iterate_corpus

HERE = pathlib.Path(__file__).parent.resolve()
RESULTS = HERE.joinpath("results")
RESULTS.mkdir(exist_ok=True, parents=True)

RESULTS_PATH = RESULTS.joinpath("medmentions_responsiveness.tsv")
RESULTS_AGG_PATH = RESULTS.joinpath("medmentions_responsiveness_aggregated.tsv")
RESULTS_AGG_TEX_PATH = RESULTS.joinpath("medmentions_responsiveness_aggregated.tex")
FIG_PATH = RESULTS.joinpath("medmentions_responsiveness.svg")
FIG_PDF_PATH = RESULTS.joinpath("medmentions_responsiveness.pdf")


def ground_package(text, **_kwargs):
    return gilda.ground(text)


def ground_package_context(text, context):
    return gilda.ground(text, context=context)


def ground_app_local(text, **_kwargs):
    return requests.post("http://localhost:8001/ground", json={"text": text}).json()


def ground_app_local_context(text, context):
    return requests.post(
        "http://localhost:8001/ground", json={"text": text, "context": context}
    ).json()


def ground_app_remote(text, **_kwargs):
    return requests.post(
        "http://grounding.indra.bio/ground", json={"text": text}
    ).json()


def ground_app_remote_context(text, context):
    return requests.post(
        "http://grounding.indra.bio/ground", json={"text": text, "context": context}
    ).json()


#: A list of benchmarks to run with three columns:
#:  type, uses context, function
FUNCTIONS = [
    ("Python package", False, ground_package),
    ("Python package", True, ground_package_context),
    ("Local web app", False, ground_app_local),
    ("Local web app", True, ground_app_local_context),
    ("Public web app", False, ground_app_remote),
    ("Public web app", True, ground_app_remote_context),
]


def run_trial(
    *, trials, corpus, func, desc: Optional[str] = None, chunk: Optional[int] = None
):
    rv = []
    outer_it = trange(trials, desc=desc)
    for trial in outer_it:
        random.shuffle(corpus)
        test_corpus = corpus[:chunk] if chunk else corpus
        inner_it = tqdm(test_corpus, desc="Examples", leave=False)
        for document_id, abstract, umls_id, text, start, end, types in inner_it:
            with logging_redirect_tqdm():
                start = time.time()
                matches = func(text, context=abstract)
                rv.append((trial, len(matches), time.time() - start))
    return rv


def build(trials: int, chunk: Optional[int] = None) -> pd.DataFrame:
    click.secho("Preparing MedMentions corpus")
    corpus = list(iterate_corpus())

    click.secho("Warming up python grounder")
    grounder.get_grounder()

    click.secho("Warming up local api grounder")
    ground_app_local_context("ER", context="Calcium is released from the ER.")

    click.secho("Warming up remote api grounder")
    ground_app_remote_context("ER", context="Calcium is released from the ER.")

    rows = []
    for tag, uses_context, func in FUNCTIONS:
        rv = run_trial(
            trials=trials,
            chunk=chunk,
            corpus=corpus,
            func=func,
            desc=f"{tag}{' with context' if uses_context else ''} trial",
        )
        rows.extend((tag, uses_context, *row) for row in rv)
    df = pd.DataFrame(rows, columns=["type", "context", "trial", "matches", "duration"])
    return df


@click.command()
@click.option("--trials", type=int, default=2, show_default=True)
@click.option(
    "--chunk",
    type=int,
    help="Subsample size from full corpus. Defaults to full corpus if not given.",
)
@verbose_option
@force_option
def main(trials: int, chunk: Optional[int], force: bool):
    if RESULTS_PATH.is_file() and not force:
        df = pd.read_csv(RESULTS_PATH, sep="\t")
    else:
        df = build(trials=trials, chunk=chunk)
        df.to_csv(RESULTS_PATH, sep="\t", index=False)

    # convert from seconds/response to responses/second
    df["duration"] = df["duration"].map(lambda x: 1 / x)

    _grouped = df[["type", "context", "duration"]].groupby(["type", "context"])
    agg_mean_df = _grouped.mean()
    agg_mean_df.rename(columns={"duration": "duration_mean"}, inplace=True)
    agg_std_df = _grouped.std()
    agg_std_df.rename(columns={"duration": "duration_std"}, inplace=True)
    agg_df = pd.merge(agg_mean_df, agg_std_df, left_index=True, right_index=True)
    agg_df = agg_df.round(1)
    agg_df.to_csv(RESULTS_AGG_PATH, sep="\t")
    agg_df.to_latex(
        RESULTS_AGG_TEX_PATH,
        label="tab:medmentions-responsiveness-benchmark",
        caption=dedent(
            f"""\
        Benchmarking of the responsiveness of the Gilda service when running synchronously
        through its Python package, when run locally as a web service, and when run remotely
        as a web service. Each scenario was also tested with and without context added.
        The Python usage had the fastest time due to the lack of overhead from
        network communication. The local web service performed better than the remote one
        for the same reason in addition to the possibility of external users requesting at the
        same time.
    """
        ),
    )

    fig, ax = plt.subplots(figsize=(6, 3))
    sns.boxplot(data=df, y="duration", x="type", hue="context", ax=ax)
    ax.set_title("Gilda Responsiveness Benchmark on MedMentions")
    ax.set_yscale("log")
    ax.set_ylabel("Responses per Second")
    ax.set_xlabel("")
    fig.savefig(FIG_PATH)
    fig.savefig(FIG_PDF_PATH)


if __name__ == "__main__":
    main()



"""This is a helper script to edit the git patch created by
changes in grounding_cell_designer.py on the COVID-19 Disease Map XMLs.
The edited patch can then be applied and committed to that repo."""
import sys


def get_blocks(fname):
    """Get blocks of the diff that can be independently filtered."""
    with open(fname, 'r') as fh:
        lines = iter(fh.readlines())
        parts = []
        line = next(lines)
        while True:
            if line.startswith('diff --git'):
                block = [line]
                for line in lines:
                    if line.startswith('@@'):
                        break
                    block.append(line)
                parts.append(block)
            if line.startswith('@@'):
                block = [line]
                for line in lines:
                    if line.startswith('@@') or line.startswith('diff --git'):
                        break
                    block.append(line)
                parts.append(block)
            if line.startswith('\\ No newline'):
                parts[-1].append(line)
                try:
                    line = next(lines)
                except StopIteration:
                    break
            if not lines:
                break
    return parts


def filter_blocks(blocks):
    """Filter out spurious diffs caused by XML deserialization/serialization."""
    new_blocks = []
    for block in blocks:
        if any(l.startswith('-<rdf:RDF') for l in block):
            continue
        if any(l.startswith('-<math') for l in block):
            continue
        if any(l.startswith('-<sbml') for l in block):
            continue
        if any(l.startswith('-<body') for l in block):
            continue
        if any('&apos;' in l for l in block):
            continue
        new_blocks.append(block)
    return new_blocks


def dump_blocks(blocks, fname):
    """Dump filtered diffs back into a patch file."""
    with open(fname, 'w') as fh:
        for block in blocks:
            for line in block:
                fh.write(line)


if __name__ == '__main__':
    # 1. run the grounding_cell_designer.py script
    # 2. in the C19DM repo run git diff --binary > patch.diff
    # 3. run this script on patch.diff
    # 4. git apply patch_edited.diff
    patch_path = sys.argv[1]
    blocks = get_blocks(patch_path)
    blocks = filter_blocks(blocks)
    dump_blocks(blocks, patch_path[:-5] + '_edited.diff')



import os
import json
import pathlib
import logging
from datetime import datetime
from collections import defaultdict, Counter
import xml.etree.ElementTree as ET
from textwrap import dedent
from typing import List, Dict

import pystow
import pandas as pd
from tqdm import tqdm

import gilda
from gilda.ner import annotate

#from benchmarks.bioid_evaluation import fplx_members
from benchmarks.bioid_evaluation import BioIDBenchmarker

logging.getLogger('gilda.grounder').setLevel('WARNING')
logger = logging.getLogger('bioid_ner_benchmark')

# Constants
HERE = os.path.dirname(os.path.abspath(__file__))

DATA_DIR = os.path.join(HERE, 'data', 'BioIDtraining_2', 'caption_bioc')
ANNOTATIONS_PATH = os.path.join(HERE, 'data', 'BioIDtraining_2',
                                'annotations.csv')
RESULTS_DIR = os.path.join(HERE, 'results', "bioid_ner_performance",
                           gilda.__version__)
MODULE = pystow.module('gilda', 'biocreative')
URL = ('https://biocreative.bioinformatics.udel.edu/media/store/files/2017'
       '/BioIDtraining_2.tar.gz')

tqdm.pandas()

BO_MISSING_XREFS = set()


class BioIDNERBenchmarker(BioIDBenchmarker):
    def __init__(self):
        super().__init__()
        print("Instantiating benchmarker...")
        self.equivalences = self._load_equivalences()
        self.paper_level_grounding = defaultdict(set)
        self.processed_data = self.process_xml_files()  # xml files processes
        self.annotations_df = self._process_annotations_table() # csvannotations
        self.gilda_annotations_map = defaultdict(list)
        self.annotations_count = 0
        self.counts_table = None
        self.precision_recall = None
        self.false_positives_counter = Counter()

    def process_xml_files(self):
        """Extract relevant information from XML files."""
        print("Extracting information from XML files...")
        data = []
        total_annotations = 0
        for filename in os.listdir(DATA_DIR):
            if filename.endswith('.xml'):
                filepath = os.path.join(DATA_DIR, filename)
                tree = ET.parse(filepath)
                root = tree.getroot()
                for document in root.findall('.//document'):
                    doc_id_full = document.find('.//id').text.strip()
                    # Split the full ID to get don_article and figure
                    don_article, figure = doc_id_full.split(' ',1)
                    don_article = don_article
                    for passage in document.findall('.//passage'):
                        offset = int(passage.find('.//offset').text)
                        text = passage.find('.//text').text
                        annotations = []
                        for annotation in passage.findall('.//annotation'):
                            annot_id = annotation.get('id')
                            annot_text = annotation.find('.//text').text
                            annot_type = annotation.find(
                                './/infon[@key="type"]').text
                            annot_offset = int(
                                annotation.find('.//location').attrib['offset'])
                            annot_length = int(
                                annotation.find('.//location').attrib['length'])
                            annotations.append({
                                'annot_id': annot_id,
                                'annot_text': annot_text,
                                'annot_type': annot_type,
                                'annot_offset': annot_offset,
                                'annot_length': annot_length,
                            })
                            total_annotations += 1
                        data.append({
                            'doc_id': don_article,
                            'figure': figure,
                            'offset': offset,
                            'text': text,
                            'annotations': annotations,
                        })
        print(f"Total annotations in XML files: {total_annotations}")
        self.annotations_count = total_annotations
        print("Finished extracting information from XML files.")
        return pd.DataFrame(data)

    def _load_equivalences(self) -> Dict[str, List[str]]:
        with open(os.path.join(HERE, 'data', 'equivalences.json')) as f:
                equivalences = json.load(f)
        return equivalences

    def _process_annotations_table(self):
        """Extract relevant information from annotations table. Modified for
        NER. Overrides the super method."""
        print("Extracting information from annotations table...")
        df = MODULE.ensure_tar_df(
            url=URL,
            inner_path='BioIDtraining_2/annotations.csv',
            read_csv_kwargs=dict(sep=',', low_memory=False),
        )
        # Split entries with multiple groundings then normalize ids
        df.loc[:, 'obj'] = df['obj'].apply(self._normalize_ids)
        # Add synonyms of gold standard groundings to help match more things
        df.loc[:, 'obj_synonyms'] = df['obj'].apply(self.get_synonym_set)
        # Create column for entity type
        df.loc[:, 'entity_type'] = df.apply(self._get_entity_type_helper,
                                            axis=1)
        processed_data = df[['text', 'obj', 'obj_synonyms', 'entity_type',
                             'don_article', 'figure', 'annot id', 'first left',
                             'last right']]
        print("%d rows in processed annotations table." % len(processed_data))
        processed_data = processed_data[processed_data.entity_type
                                        != 'unknown']
        print("%d rows in annotations table without unknowns." %
              len(processed_data))
        for don_article, text, synonyms in df[['don_article', 'text',
                                               'obj_synonyms']].values:
            self.paper_level_grounding[don_article, text].update(synonyms)
        return processed_data

    def annotate_entities_with_gilda(self):
        """Performs NER on the XML files using gilda.annotate()"""
        print("Annotating corpus with Gilda...")

        total_gilda_annotations = 0
        for _, item in tqdm(self.processed_data.iterrows(),
                            total=self.processed_data.shape[0],
                            desc="Annotating with Gilda"):
            doc_id = item['doc_id']
            figure = item['figure']
            text = item['text']

            # Get the full text for the paper-level disambiguation
            full_text = self._get_plaintext(doc_id)

            gilda_annotations = annotate(text, context_text=full_text)

            for annotation in gilda_annotations:
                total_gilda_annotations += 1

                self.gilda_annotations_map[(doc_id, figure)].append(annotation)

        tqdm.write("Finished annotating corpus with Gilda...")
        tqdm.write(f"Total Gilda annotations: {total_gilda_annotations}")

    def evaluate_gilda_performance(self):
        """Calculates precision, recall, and F1"""
        print("Evaluating performance...")

        metrics = {
            'all_matches': {'tp': 0, 'fp': 0, 'fn': 0},
            'top_match': {'tp': 0, 'fp': 0, 'fn': 0}
        }



        ref_dict = defaultdict(list)
        for _, row in self.annotations_df.iterrows():
            key = (str(row['don_article']), row['figure'], row['text'],
                   row['first left'], row['last right'])
            ref_dict[key].append((set(row['obj']), row['obj_synonyms']))

        for (doc_id, figure), annotations in (
                tqdm(self.gilda_annotations_map.items(),
                     desc="Evaluating Annotations")):
            for annotation in annotations:
                key = (doc_id, figure, annotation.text, annotation.start,
                       annotation.end)
                matching_refs = ref_dict.get(key, [])

                match_found = False
                for i, scored_match in enumerate(annotation.matches):
                    curie = f"{scored_match.term.db}:{scored_match.term.id}"

                    for original_curies, synonyms in matching_refs:
                        if curie in original_curies or curie in synonyms:
                            metrics['all_matches']['tp'] += 1
                            if i == 0:  # Top match
                                metrics['top_match']['tp'] += 1
                            match_found = True
                            break

                    if match_found:
                        break

                if not match_found:
                    metrics['all_matches']['fp'] += 1
                    self.false_positives_counter[annotation.text] += 1
                    if annotation.matches:  # Check if there are any matches
                        metrics['top_match']['fp'] += 1

        # False negative calculation using ref dict
        for key, refs in tqdm(ref_dict.items(),
                              desc="Calculating False Negatives"):
            doc_id, figure = key[0], key[1]
            gilda_annotations = self.gilda_annotations_map.get((doc_id, figure),
                                                               [])
            for original_curies, synonyms in refs:
                match_found = any(
                    ann.text == key[2] and
                    ann.start == key[3] and
                    ann.end == key[4] and
                    any(f"{match.term.db}:{match.term.id}" in original_curies or
                        f"{match.term.db}:{match.term.id}" in synonyms
                        for match in ann.matches)
                    for ann in gilda_annotations
                )

                if not match_found:
                    metrics['all_matches']['fn'] += 1
                    metrics['top_match']['fn'] += 1

        results = {}
        for match_type, counts in metrics.items():
            precision = counts['tp'] / (counts['tp'] + counts['fp']) \
                if ((counts['tp'] + counts['fp']) > 0) else 0

            recall = counts['tp'] / (counts['tp'] + counts['fn']) \
                if (counts['tp'] + counts['fn']) > 0 else 0

            f1 = 2 * (precision * recall) / (precision + recall) \
                if (precision + recall) > 0 else 0

            results[match_type] = {
                'precision': precision,
                'recall': recall,
                'f1': f1
            }

        counts_table = pd.DataFrame([
            {
                'Match Type': 'All Matches',
                'True Positives': metrics['all_matches']['tp'],
                'False Positives': metrics['all_matches']['fp'],
                'False Negatives': metrics['all_matches']['fn']
            },
            {
                'Match Type': 'Top Match',
                'True Positives': metrics['top_match']['tp'],
                'False Positives': metrics['top_match']['fp'],
                'False Negatives': metrics['top_match']['fn']
            }
        ])

        precision_recall = pd.DataFrame([
            {
                'Match Type': 'All Matches',
                'Precision': results['all_matches']['precision'],
                'Recall': results['all_matches']['recall'],
                'F1 Score': results['all_matches']['f1']
            },
            {
                'Match Type': 'Top Match',
                'Precision': results['top_match']['precision'],
                'Recall': results['top_match']['recall'],
                'F1 Score': results['top_match']['f1']
            }
        ])

        self.counts_table = counts_table
        self.precision_recall = precision_recall

        os.makedirs(RESULTS_DIR, exist_ok=True)
        false_positives_df = pd.DataFrame(self.false_positives_counter.items(),
                                          columns=['False Positive Text',
                                                   'Count'])
        false_positives_df = false_positives_df.sort_values(by='Count',
                                                            ascending=False)
        false_positives_df.to_csv(
            os.path.join(RESULTS_DIR, 'false_positives.csv'), index=False)

        print("Finished evaluating performance...")

    def get_tables(self):
        return (self.counts_table,
                self.precision_recall,
                self.false_positives_counter)


def main(results: str = RESULTS_DIR):
    results_path = os.path.expandvars(os.path.expanduser(results))
    os.makedirs(results_path, exist_ok=True)

    benchmarker = BioIDNERBenchmarker()
    benchmarker.annotate_entities_with_gilda()
    benchmarker.evaluate_gilda_performance()
    counts, precision_recall, false_positives_counter = benchmarker.get_tables()

    print(f"Counts Table:")
    print(counts.to_markdown(index=False))
    print(f"Precision and Recall table: ")
    print(precision_recall.to_markdown(index=False))

    time = datetime.now().strftime('%y%m%d-%H%M%S')

    outname = f'benchmark_{time}'
    result_stub = pathlib.Path(results_path).joinpath(outname)

    caption0 = dedent(f"""\
        # Gilda NER Benchmarking

        Gilda: v{gilda.__version__}
        Date: {time}
        """)

    caption1 = dedent("""\
            ## Table 1

            The counts of true positives, false positives, and false negatives 
            for Gilda annotations in the corpus where only Gilda's "Top Match"
            grounding (top score grounding) returns the correct match and where
            any Gilda grounding returns a correct match.
        """)
    table1 = counts.to_markdown(index=False)

    caption2 = dedent("""\
            ## Table 2

            Precision, recall, and F1 Score values for Gilda performance where 
            Gilda's "Top Match" grounding (top score grounding) returns the 
            correct match and where any Gilda grounding returns a correct match.
        """)
    table2 = precision_recall.to_markdown(index=False)

    caption3 = dedent("""\
           ## 50 Most Common False Positive Words

           A list of 50 most common false positive annotations created by Gilda.
       """)
    top_50_false_positives = false_positives_counter.most_common(50)
    false_positives_list = '\n'.join(
        [f'- {word}: {count}' for word, count in top_50_false_positives])

    output = '\n\n'.join([
        caption0,
        caption1, table1,
        caption2, table2,
        caption3, false_positives_list
    ])

    md_path = result_stub.with_suffix(".md")
    with open(md_path, 'w') as f:
        f.write(output)

    counts.to_csv(result_stub.with_suffix(".counts.csv"), index=False)
    precision_recall.to_csv(result_stub.with_suffix(".precision_recall.csv"),
                            index=False)
    print(f'Results saved to {results_path}')


if __name__ == '__main__':
    main()



"""
Gilda implements a simple dictionary-based named entity
recognition (NER) algorithm. It can be used as follows:

>>> from gilda.ner import annotate
>>> text = "MEK phosphorylates ERK"
>>> results = annotate(text)

The results are a list of Annotation objects each of which contains:

- the `text` string matched
- a list of :class:`gilda.grounder.ScoredMatch` instances containing a sorted list of matches
  for the given text span (first one is the best match)
- the `start` position in the text string where the entity starts
- the `end` position in the text string where the entity ends


In this example, the two concepts are grounded to FamPlex entries.

>>> results[0].text, results[0].matches[0].term.get_curie(), results[0].start, results[0].end
('MEK', 'fplx:MEK', 0, 3)
>>> results[1].text, results[1].matches[0].term.get_curie(), results[1].start, results[1].end
('ERK', 'fplx:ERK', 19, 22)

If you directly look in the second part of the 4-tuple, you get a full
description of the match itself:

>>> results[0].matches[0]
ScoredMatch(Term(mek,MEK,FPLX,MEK,MEK,curated,famplex,None,None,None),\
0.9288806431663574,Match(query=mek,ref=MEK,exact=False,space_mismatch=\
False,dash_mismatches=set(),cap_combos=[('all_lower', 'all_caps')]))

BRAT
----
Gilda implements a way to output annotation in a format appropriate for the
`BRAT Rapid Annotation Tool (BRAT) <https://brat.nlplab.org/index.html>`_.

>>> from gilda.ner import get_brat
>>> from pathlib import Path
>>> brat_string = get_brat(results)
>>> Path("results.ann").write_text(brat_string)
>>> Path("results.txt").write_text(text)

For brat to work, you need to store the text in a file with
the extension ``.txt`` and the annotations in a file with the
same name but extension ``.ann``.
"""

from typing import List, Set
import os

from nltk.corpus import stopwords
from nltk.tokenize import PunktSentenceTokenizer, TreebankWordTokenizer

from gilda import get_grounder
from gilda.grounder import Annotation
from gilda.process import normalize

__all__ = [
    "annotate",
    "get_brat",
    "stop_words"
]

STOPLIST_PATH = os.path.join(os.path.dirname(__file__),'resources',
                             'ner_stoplist.txt')


def _load_stoplist() -> Set[str]:
    """Load NER stoplist from file."""
    stoplist_path = STOPLIST_PATH
    with open(stoplist_path, 'r') as file:
        stoplist = {line.strip() for line in file}
    return stoplist


core_stop_words = set(stopwords.words('english'))
stop_words = core_stop_words | _load_stoplist()


def annotate(
    text, *,
    grounder=None,
    sent_split_fun=None,
    organisms=None,
    namespaces=None,
    context_text: str = None,
) -> List[Annotation]:
    """Annotate a given text with Gilda.

    Parameters
    ----------
    text : str
        The text to be annotated.
    grounder : gilda.grounder.Grounder, optional
        The Gilda grounder to use for grounding.
    sent_split_fun : Callable[str, Iterable[Tuple[int, int]]], optional
        A function that splits the text into sentences. The default is
        :func:`nltk.tokenize.PunktSentenceTokenizer.span_tokenize`. The function
        should take a string as input and return an iterable of coordinate pairs
        corresponding to the start and end coordinates for each sentence in the
        input text.
    organisms : list[str], optional
        A list of organism names to pass to the grounder. If not provided,
        human is used.
    namespaces : List[str], optional
        A list of namespaces to pass to the grounder to restrict the matches
        to. By default, no restriction is applied.
    context_text :
        A longer span of text that serves as additional context for the text
        being annotated for disambiguation purposes.

    Returns
    -------
    List[Annotation]
        A list of Annotations where each contains as attributes
        the text span that was matched, the list of ScoredMatches, and the
        start and end character offsets of the text span.
    """
    if grounder is None:
        grounder = get_grounder()
    if sent_split_fun is None:
        sent_tokenizer = PunktSentenceTokenizer()
        sent_split_fun = sent_tokenizer.span_tokenize
    # Get sentences
    sentence_coords = sent_split_fun(text)
    text_coord = 0
    annotations = []
    word_tokenizer = TreebankWordTokenizer()
    # FIXME: a custom sentence split function can be inconsistent
    # with the coordinates being used here which come from NLTK
    for sent_start, sent_end in sentence_coords:
        sentence = text[sent_start:sent_end]
        # FIXME: one rare corner case is named entities with single quotes
        # in them which get tokenized in a weird way
        raw_word_coords = \
            list(word_tokenizer.span_tokenize(sentence.rstrip('.')))
        raw_words = [sentence[start:end] for start, end in raw_word_coords]
        text_coord += len(sentence) + 1
        words = [normalize(w) for w in raw_words]
        skip_until = 0
        for idx, word in enumerate(words):
            if idx < skip_until:
                continue
            if word in core_stop_words:
                continue
            spans = grounder.prefix_index.get(word, set())
            if not spans:
                continue

            # Only consider spans that are within the sentence
            applicable_spans = {span for span in spans
                                if idx + span <= len(words)}

            # Find the largest matching span
            for span in sorted(applicable_spans, reverse=True):
                # If the span we are looking at is a single word
                # and it is a stopword then we skip this. This ensures
                # that we don't skip longer spans of which the stopword
                # is only a part.
                if span == 1 and word in stop_words:
                    continue
                # We have to reconstruct a text span while adding spaces
                # where needed
                raw_span = ''
                for rw, c in zip(raw_words[idx:idx+span],
                                 raw_word_coords[idx:idx+span]):
                    # Figure out if we need a space before this word, then
                    # append the word.
                    spaces = ' ' * (c[0] - len(raw_span) -
                                    raw_word_coords[idx][0])
                    raw_span += spaces + rw
                # If span is a single character, we don't want to consider it
                if len(raw_span) <= 1:
                    continue
                context = text if context_text is None else context_text
                matches = grounder.ground(raw_span,
                                          context=context,
                                          organisms=organisms,
                                          namespaces=namespaces)
                if matches:
                    start_coord = sent_start + raw_word_coords[idx][0]
                    end_coord = sent_start + raw_word_coords[idx+span-1][1]
                    annotations.append(Annotation(
                        raw_span, matches, start_coord, end_coord
                    ))

                    skip_until = idx + span
                    break
    return annotations


def get_brat(annotations, entity_type="Entity", ix_offset=1, include_text=True):
    """Return brat-formatted annotation strings for the given entities.

    Parameters
    ----------
    annotations : list[Annotation]
        A list of named entity annotations in the text.
    entity_type : str, optional
        The brat entity type to use for the annotations. The default is
        'Entity'. This is useful for differentiating between annotations in
        the same text extracted from different reading systems.
    ix_offset : int, optional
        The index offset to use for the brat annotations. The default is 1.
    include_text : bool, optional
        Whether to include the text of the entity in the brat annotations.
        The default is True. If not provided, the text that matches the span
        will be written to the annotation file.

    Returns
    -------
    str
        A string containing the brat-formatted annotations.
    """
    brat = []
    ix_offset = max(1, ix_offset)
    for idx, annotation in enumerate(annotations, ix_offset):
        curie = annotation.matches[0].term.get_curie()
        if entity_type != "Entity":
            curie += f"; Reading system: {entity_type}"
        row = f'T{idx}\t{entity_type} {annotation.start} {annotation.end}' + (
            f'\t{annotation.text}' if include_text else ''
        )
        brat.append(row)
        row = f'#{idx}\tAnnotatorNotes T{idx}\t{curie}'
        brat.append(row)
    return '\n'.join(brat) + '\n'



"""Test the API in the Gilda app."""

import unittest
from typing import ClassVar

import flask

from gilda.app.app import get_app


class TestApp(unittest.TestCase):
    """A test case for the Gilda Flask application."""

    app: ClassVar[flask.Flask]

    @classmethod
    def setUpClass(cls) -> None:
        cls.app = get_app()

    def test_get_home(self):
        """Test the GET response on the home page."""
        with self.app.test_client() as client:
            res = client.get("/?text=Raf1")
            self.assertNotEqual(
                302,
                res.status_code,
                msg="Should not receive a redirect, this probably means the UI isn't mounted properly",
            )
            self.assert_raf1_ui(res)

    def test_post_home(self):
        """Test the POST response on the home page."""
        with self.app.test_client() as client:
            res = client.post("/", json={"text": "Raf1"})
            self.assertNotEqual(
                302,
                res.status_code,
                msg="Should not receive a redirect, this probably means the UI isn't mounted properly",
            )
            self.assert_raf1_ui(res)

    def assert_raf1_ui(self, res) -> None:
        text = res.data.decode()
        self.assertEqual(200, res.status_code)
        self.assertIn('RAF1', text)
        self.assertIn('0.9998', text)
        self.assertIn('RNASE3', text)
        self.assertIn('0.5024', text)

    def test_post_grounding(self):
        """Test the POST response with text."""
        with self.app.test_client() as client:
            res = client.post("/ground")
            self.assertIn("message", res.json)

            res = client.post("/ground", json={"text": "Raf1"})
            self.assertIsInstance(res.json, list)
            self.assert_found(res.json, "HGNC", "9829")

            res = client.post("/ground", json={"text": "Raf1", "organisms": ["9606", "10090"]})
            self.assertIsInstance(res.json, list)
            self.assert_found(res.json, "HGNC", "9829")

            res = client.post("/ground", json={"text": "Raf1", "organisms": ["10090", "9606"]})
            self.assertIsInstance(res.json, list)
            self.assert_found(res.json, "UP", "Q99N57")

    def test_get_names(self):
        with self.app.test_client() as client:
            res = client.post('/names', json={"db": "FPLX", "id": "ERK"})
            self.assertIsInstance(res.json, list)
            self.assertIn('ERK 1/2', res.json)

    def test_get_models(self):
        with self.app.test_client() as client:
            res = client.get('/models')
            self.assertIsInstance(res.json, list)
            self.assertIn('ABC1', res.json)

    def assert_found(self, matches, prefix: str, identifier: str) -> None:
        match_curies = {
            (match['term']['db'], match['term']['id'])
            for match in matches
        }
        self.assertIn((prefix, identifier), match_curies)



"""This is a script that helps deal with diffs of benchmarks."""
import os
from pathlib import Path
from textwrap import dedent

import click
import pandas as pd
from matplotlib import pyplot as plt
from matplotlib_venn import venn2

HERE = Path(__file__).parent.resolve()
RESULTS = HERE.joinpath('results', 'bioid_performance')


@click.command()
@click.option("--reference")
@click.option("--comparison")
@click.option("--key", default="exists_correct")
def main(reference: str, comparison: str, key: str):
    df1 = pd.read_json(reference, orient="record")
    df2 = pd.read_json(comparison, orient="record")
    reference_base = os.path.splitext(os.path.basename(reference))[0]
    comparison_base = os.path.splitext(os.path.basename(comparison))[0]
    output = RESULTS.joinpath(f"{reference_base}_{comparison_base}", key)
    output.mkdir(exist_ok=True, parents=True)

    fig, ax = plt.subplots()
    venn2(
        [
            set(df1[df1[key]].index),
            set(df2[df2[key]].index),
        ],
        [
            reference_base,
            comparison_base,
        ],
        ax=ax,
    )
    fig.savefig(output.joinpath("comparison.svg"))

    # Here are the texts that grounded in the new version but not the old
    new_idx = df2[key] & ~df1[key]
    newly_grounded = df2[new_idx].text.value_counts()
    newly_grounded.to_csv(output.joinpath("improvements.tsv"), sep='\t')
    new_unique_count = df2[new_idx].text.nunique()

    # Here are the texts that grounded in the old version but not the new
    regressions_idx = df1[key] & ~df2[key]
    regressions = df2[regressions_idx].text.value_counts()
    regressions.to_csv(output.joinpath("regressions.tsv"), sep='\t')
    regression_unique_count = df2[regressions_idx].text.nunique()

    # Here are the texts that neither versions groudned right
    misses_idx = ~df1[key] & ~df2[key]
    misses = df2[misses_idx].text.value_counts()
    misses.to_csv(output.joinpath("misses.tsv"), sep='\t')
    misses_unique_count = df2[misses_idx].text.nunique()

    hits_idx = df1[key] & df2[key]
    hits_unique_count = df2[hits_idx].text.nunique()

    print(dedent(f"""\
    Analysis of "{key}":
    
    {new_idx.sum():,} success rows ({new_unique_count:,} unique) in {comparison_base} but not {reference_base}
    {regressions_idx.sum():,} success rows ({regression_unique_count:,} unique) in {reference_base} but not {comparison_base}
    {misses.sum():,} rows missed ({misses_unique_count:,} unique) in both
    {hits_idx.sum():,} successes rows ({hits_unique_count:,} unique) in both
    """))


if __name__ == '__main__':
    main()



"""This script can process a set of Cell Designer XML files that are
part of the COVID-19 Disease Map project, detect species with missing
grounding, ground these based on their string name with Gilda, and
serialize the changes back into XML files."""

import re
import itertools
from collections import defaultdict
from pathlib import Path
from xml.etree import ElementTree as ET

import click
from indra.databases import identifiers
from tqdm import tqdm

import gilda

rdf_str = (
    b'<rdf:RDF '
    b'xmlns:bqbiol="http://biomodels.net/biology-qualifiers/" '
    b'xmlns:bqmodel="http://biomodels.net/model-qualifiers/" '
    b'xmlns:dc="http://purl.org/dc/elements/1.1/" '
    b'xmlns:dcterms="http://purl.org/dc/terms/" '
    b'xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#" '
    b'xmlns:vCard="http://www.w3.org/2001/vcard-rdf/3.0#">'
)

namespaces = {'sbml': 'http://www.sbml.org/sbml/level2/version4',
              'celldesigner': 'http://www.sbml.org/2001/ns/celldesigner',
              'rdf': 'http://www.w3.org/1999/02/22-rdf-syntax-ns#',
              'bqbiol': 'http://biomodels.net/biology-qualifiers/',
              'bqmodel': 'http://biomodels.net/model-qualifiers/',
              'vCard': 'http://www.w3.org/2001/vcard-rdf/3.0#',
              'dc': 'http://purl.org/dc/elements/1.1/',
              'xhtml': 'http://www.w3.org/1999/xhtml'}


irrelevant_grounding_ns = {'pubmed', 'taxonomy', 'doi', 'wikipathways', 'pdb',
                           'intact', 'biogrid', 'pmc'}
relevant_grounding_ns = {'ncbiprotein', 'ncbigene', 'uniprot', 'obo.go',
                         'obo.chebi', 'pubchem.compound', 'pubchem.substance',
                         'hgnc', 'hgnc.symbol', 'mesh', 'interpro',
                         'refseq', 'ensembl', 'ec-code', 'brenda',
                         'kegg.compound', 'drugbank'}
irrelevant_classes = {'DEGRADED'}
grounding_stats = {'ngrounding': 0}


def register_all_namespaces(fname):
    namespaces = dict([node for _, node in
                       ET.iterparse(fname, events=['start-ns'])])
    for ns in namespaces:
        ET.register_namespace(ns, namespaces[ns])


def get_protein_reference(species):
    protein_ref = species.find('sbml:annotation/'
                               'celldesigner:extension/'
                               'celldesigner:speciesIdentity/'
                               'celldesigner:proteinReference',
                               namespaces=namespaces)
    if protein_ref is None:
        protein_ref = species.find('celldesigner:annotation/'
                                   'celldesigner:speciesIdentity/'
                                   'celldesigner:proteinReference',
                                   namespaces=namespaces)
        if protein_ref is None:
            protein_ref = species.find('celldesigner:extension/'
                                       'celldesigner:speciesIdentity/'
                                       'celldesigner:proteinReference',
                                       namespaces=namespaces)

    if protein_ref is not None:
        protein_ref_id = protein_ref.text
        return protein_ref_id


def add_groundings(et):
    all_species = et.findall('sbml:model/sbml:listOfSpecies/sbml:species',
                             namespaces=namespaces)
    print('%d species found with first method' % len(all_species))
    all_species += et.findall('sbml:model/sbml:annotation/'
                              'celldesigner:extension/'
                              'celldesigner:listOfIncludedSpecies/'
                              'celldesigner:species',
                              namespaces=namespaces)
    print('%d species found after second method' % len(all_species))
    # This is to collect protein reference-based groundings first
    groundings_for_prot_ref = defaultdict(set)
    species_for_prot_ref = defaultdict(set)
    for species in all_species:
        protein_ref_id = get_protein_reference(species)
        if protein_ref_id:
            existing_grounding = get_existing_grounding(species)
            groundings_for_prot_ref[protein_ref_id] |= \
                set(existing_grounding) & relevant_grounding_ns
            species_for_prot_ref[protein_ref_id].add(species.attrib['id'])
    for species in tqdm(all_species, desc='Species', leave=False):
        # Skip if we have a grounding via the protein reference, indirectly
        protein_ref_id = get_protein_reference(species)
        if protein_ref_id and groundings_for_prot_ref.get(protein_ref_id):
            continue
        existing_grounding = get_existing_grounding(species)
        # Important: this is where we decide if we will add any grounding.
        # Here we skip this species if it has any relevant grounding
        if set(existing_grounding) & relevant_grounding_ns:
            continue
        class_tag = species.find(
            'sbml:annotation/celldesigner:extension/'
            'celldesigner:speciesIdentity/celldesigner:class',
            namespaces=namespaces)
        if class_tag is not None:
            entity_class = class_tag.text
            if entity_class in irrelevant_classes:
                continue
        name = species.attrib.get('name')
        tqdm.write(name)
        tqdm.write(entity_class)
        matches = gilda.ground(name)
        if matches:
            species = add_grounding_element(species, entity_class,
                                            matches[0].term.db,
                                            matches[0].term.id)
            if species:
                tqdm.write(' '.join((name, matches[0].term.db, matches[0].term.id,
                      matches[0].term.entry_name)))
                grounding_stats['ngrounding'] += 1
        tqdm.write('---')
    return et


def get_existing_grounding(species):
    # Others: isHomologTo
    bqbio_tags = ['isDescribedBy', 'isEncodedBy', 'is', 'encodes',
                  'occursIn']
    prefixes = ['sbml:annotation', 'celldesigner:notes/xhtml:html/xhtml:body']

    groundings = defaultdict(list)
    for prefix, tag in itertools.product(prefixes, bqbio_tags):
        grounding_elements = \
            species.findall('%s/rdf:RDF/rdf:Description/'
                            'bqbiol:%s/rdf:Bag/rdf:li' % (prefix, tag),
                            namespaces=namespaces)
        for element in grounding_elements:
            urn = element.attrib['{http://www.w3.org/1999/02/22-rdf-syntax-ns#}'
                                 'resource']
            match = re.match(r'urn:miriam:([^:]+):(.+)', urn)
            if not match:
                tqdm.write(f'Unmatched urn: {urn}')
                continue
            else:
                db_ns, db_id = match.groups()
                groundings[db_ns].append(db_id)
    return groundings


def add_grounding_element(species, entity_class, db_ns, db_id):
    # For genes, if we're grounding a protein, we make the encoding aspect
    # explicit
    if entity_class == 'PROTEIN' and db_ns == 'HGNC':
        bqbiol_tag = 'bqbiol:isEncodedBy'
    # In case a protein is grounded to CHEBI, it's typically a problem, we
    # skip these
    elif entity_class == 'PROTEIN' and db_ns == 'CHEBI':
        return
    else:
        bqbiol_tag = 'bqbiol:is'

    tag_sequence = [
        'sbml:annotation',
        'rdf:RDF',
        'rdf:Description',
        bqbiol_tag,
        'rdf:Bag',
    ]

    identifiers_ns = identifiers.get_identifiers_ns(db_ns)
    grounding_str = 'urn:miriam:%s:%s' % (identifiers_ns, db_id)

    root = species
    for tag in tag_sequence:
        element = root.find(tag, namespaces=namespaces)
        if element is not None:
            root = element
        else:
            new_element = ET.Element(tag)
            new_element.text = '\n'
            new_element.tail = '\n'
            root.append(new_element)
            root = new_element

    li = ET.Element('rdf:li', attrib={'rdf:resource': grounding_str})
    root.append(li)
    return species


def dump(et, fname):
    xml_str = ET.tostring(et.getroot(), xml_declaration=False,
                          encoding='UTF-8')
    xml_str = b'<?xml version="1.0" encoding="UTF-8"?>\n' + xml_str
    xml_str = xml_str.replace(b'ns0:', b'')
    xml_str = xml_str.replace(b'ns1:', b'')
    xml_str = xml_str.replace(b' />', b'/>')
    xml_str = xml_str.replace(b'<html>',
                              b'<html xmlns="http://www.w3.org/1999/xhtml">')
    xml_str = xml_str.replace(b'<rdf:RDF>', rdf_str)
    xml_str = xml_str.replace(b'xmlns:ns0="http://www.sbml.org/sbml/level2/version4"',
                              b'')
    xml_str = xml_str.replace(b'xmlns:ns1="http://www.w3.org/1999/xhtml"', b'')
    with open(fname, 'wb') as fh:
        fh.write(xml_str)


@click.command()
@click.argument('directory', type=Path)
def main(directory: Path):
    """Run grounding on the directory for the COVID 19 Disease Maps repository."""
    stable_xmls = list(directory.resolve().rglob('*_stable.xml'))
    it = tqdm(stable_xmls, desc='Grounding')
    for stable_xml in it:
        it.set_postfix(file=stable_xml.name)
        register_all_namespaces(stable_xml)
        et = ET.parse(stable_xml)
        et = add_groundings(et)
        dump(et, stable_xml)


if __name__ == '__main__':
    gilda.ground('x')
    main()



from gilda.tests import appreq
from gilda.api import *
from gilda.term import Term


def test_api_ground():
    scores = ground('kras')
    assert appreq(scores[0].score, 0.9845), scores
    scores = ground('ROS', 'reactive oxygen')
    assert scores[0].term.db == 'MESH', scores
    assert scores[0].url == 'https://bioregistry.io/mesh:D017382'


def test_get_models():
    models = get_models()
    assert len(models) > 500
    assert 'STK1' in models


def test_get_names():
    names = get_names('HGNC', '6407')
    assert len(names) > 5, names
    assert 'K-Ras' in names


def test_organisms():
    # Default human gene match
    matches1 = ground('SMN1')
    assert len(matches1) == 1
    assert matches1[0].term.db == 'HGNC'
    assert matches1[0].term.id == '11117'
    # Prioritize human gene match
    matches2 = ground('SMN1', organisms=['9606', '10090'])
    assert len(matches2) == 1
    assert matches2[0].term.db == 'HGNC'
    assert matches2[0].term.id == '11117'
    # Prioritize mouse, SMN is grounded correctly
    matches3 = ground('SMN', organisms=['10090', '9606'])
    assert len(matches3) == 2, matches3
    assert matches3[0].term.db == 'UP'
    assert matches3[0].term.id == 'P63163'
    # Here we use SMN again but prioritize human and get three bad groundings
    matches4 = ground('SMN', organisms=['9606', '10090'])
    assert len(matches4) == 2, matches4
    assert all(m.term.organism == '9606' for m in matches4)
    # Finally we try grounding SMN1 with mouse prioritized, don't find a match
    # and end up with the human gene grounding
    matches5 = ground('TDRD16A', organisms=['10090', '9606'])
    assert len(matches5) == 1, matches5
    assert matches5[0].term.db == 'HGNC', matches5
    assert matches5[0].term.id == '11117', matches5


def test_make_grounder():
    grounder = make_grounder([
        Term('a', 'A', 'X', '1', 'A', 'name', 'test'),
        Term('b', 'B', 'X', '2', 'B', 'name', 'test')
    ])
    assert grounder.ground('a')
    assert not grounder.ground('x')



import copy
import pandas
import itertools
from indra.databases import chebi_client
from gilda import ground
from tqdm import tqdm

url = ('https://raw.githubusercontent.com/sorgerlab/famplex_paper/master/'
       'step4_stmt_entity_stats/test_agents_with_fplx_sample_curated.csv')

# Here we do some manual annotation to add expected correct groundings
# for some entries that were incorrect or ungrounded in the reference.
# This will allow us to compare more of the groundings to known correct
# groundings
correct_assertions = {'Stat': {'FPLX': 'STAT'},
                      'Thioredoxin Reductase': {'FPLX': 'TXNRD'},
                      'Thrombin': {'HGNC': '3535'},
                      'Aminopeptidases': {'MESH': 'D000626'},
                      'NF-AT proteins': {'MESH': 'D050778'},
                      'LTbetaR': {'HGNC': '6718'},
                      'RNAi': {'MESH': 'D034622', 'GO': 'GO:0016441'},
                      'Chaetocin': {'CHEBI': 'CHEBI:68747'},
                      'BAY11-7082': {'CHEBI': 'CHEBI:85928'},
                      'Toll-like receptors': {'MESH': 'D051193'},
                      'hippocalcin': {'HGNC': '5144'},
                      'alanine': {'CHEBI': 'CHEBI:16449'},
                      'H3K4me3': {'CHEBI': 'CHEBI:85043'},
                      'RXRalpha': {'HGNC': '10477'},
                      'chemokine receptors': {'MESH': 'D019707'},
                      'laminin': {'MESH': 'D007797'},
                      'Tiron': {'CHEBI': 'CHEBI:9607'},
                      'factor VII': {'HGNC': '3544'},
                      'CD73': {'HGNC': '8021'},
                      'SERM': {'MESH': 'D020845'},
                      'angiotensin II': {'CHEBI': 'CHEBI:48432'},
                      'NAPA': {'HGNC': '7641'},
                      'IFN-beta': {'FPLX': 'IFNB'},
                      'cyclooxygenase-2': {'HGNC': '9605'},
                      'substance P': {'CHEBI': 'CHEBI:80308'},
                      'progestin': {'CHEBI': 'CHEBI:59826'},
                      'MiR-125a': {'HGNC': '31505'},
                      'ROS': {'MESH': 'D017382'},
                      'PP5': {'HGNC': '9322'},
                      'aminopeptidases': {'MESH': 'D000626'},
                      'IMP1': {'HGNC': '28866'},
                      '293T': {'EFO': '0001082'},
                      'GR': {'HGNC': '7978'},
                      'integrin alpha': {'FPLX': 'ITGA'},
                      'DC': {'MESH': 'D003713'},
                      'BMD': {'MESH': 'D015519'},
                      'angina': {'MESH': 'D000787', 'EFO': '0003913'}}


incorrect_assertions = {'IGF': {'HGNC': '5464'},
                        'DHM': {'CHEBI': 'CHEBI:71175'},
                        'SM': {'CHEBI': 'CHEBI:17076'},
                        'BMD': {'HGNC': ['2928', '12703']},
                        'DC': {'HGNC': '2714'},
                        'ARs': {'HGNC': '644'},
                        'BA': {'CHEBI': 'CHEBI:32594'},
                        'HP1': {'HGNC': '1555'},
                        'PRF': {'UP': 'Q6P3D7'},
                        'CUL4': {'UP': 'Q17392'},
                        'MEKK3': {'GO': 'GO:0004709'},
                        'Hrs': {'GO': 'GO:0000725'},
                        'thioredoxin-1': {'UP': 'P47938'},
                        'alpha4': {'HGNC': '10809'},
                        'NT': {'HGNC': '17941'},
                        'IMP1': {'HGNC': '16435'}}


def process_fplx_groundings(df):
    groundings = []
    # Iterate over the rows of the curation table and extract groundings
    for _, row in tqdm(df.iterrows(), desc='Processing groundings'):
        if pandas.isnull(row['Grounding']):
            break
        # Here we get the original entity text, its type, and the
        # correct/incorrect curation
        grounding = {
            'text': row['Text'],
            'entity_type': row['EntityType'],
            'db_refs': {},
            'correct': bool(int(row['Grounding'])),
            'context': row['Sentence']
            }
        # We then extract the grounding (up to 3) that were considered
        # for the curation
        for i in [1, 2, 3]:
            if not pandas.isnull(row['DB_Ns%d' % i]):
                grounding['db_refs'][row['DB_Ns%d' % i]] = row['DB_Id%d' % i]
        # We standardize some of the grounding entries to match up with
        # Gilda's format
        for k, v in copy.deepcopy(grounding['db_refs']).items():
            # Strip off extra GO prefixes
            if v.startswith('GO:GO'):
                grounding['db_refs'][k] = v[3:]
            # Get CHEBI IDs from PUBCHEM
            if k == 'PUBCHEM':
                chebi_id = chebi_client.get_chebi_id_from_pubchem(v)
                if chebi_id:
                    grounding['db_refs']['CHEBI'] = chebi_id
        groundings.append(grounding)
    return groundings


def evaluate_old_grounding(grounding):
    """Return status of old grounding."""
    if not grounding['db_refs']:
        return 'ungrounded'
    elif not grounding['correct']:
        return 'incorrect'
    else:
        return 'correct'


def evaluate_new_grounding(grounding, term):
    """Return status of new grounding by comparing to old grounding."""
    if grounding['text'] in correct_assertions:
        if term.id == correct_assertions[grounding['text']].get(term.db):
            return 'correct'
    if grounding['text'] in incorrect_assertions:
        incorrect_val = incorrect_assertions[grounding['text']].get(term.db)
        if isinstance(incorrect_val, str):
            if term.id == incorrect_val:
                return 'incorrect'
        elif isinstance(incorrect_val, list):
            if term.id in incorrect_val:
                return 'incorrect'

    if not grounding['correct']:
        # If the grounding matches one of the known incorrect ones
        if grounding['db_refs'].get(term.db) == term.id:
            return 'incorrect'
    else:
        # If the grounding matches one of the known correct ones
        if grounding['db_refs'].get(term.db) == term.id:
            return 'correct'
    return 'unknown'


def make_comparison(groundings, use_disamb=True):
    # Generate an initial comparison matrix
    # This dict contains counts of all the possible relationships between
    # the reference grounding and the one produced by Gilda
    comparison = {'%s_%s' % (a, b): [] for a, b in
                  itertools.product(['ungrounded', 'correct', 'incorrect'],
                                    ['ungrounded', 'unknown', 'correct',
                                     'incorrect'])}

    # Now iterate over all the old groundings, get the new one, and build up
    # the values in the comparison matrix
    for idx, grounding in enumerate(tqdm(groundings, desc='Making comparison')):
        old_eval = evaluate_old_grounding(grounding)
        # Send grounding requests
        context = grounding['context'] if use_disamb else None
        matches = ground(text=grounding['text'],
                         context=context)
        if not matches:
            comparison['%s_ungrounded' % old_eval].append((idx, grounding,
                                                           None))
            continue
        term = matches[0].term
        new_eval = evaluate_new_grounding(grounding, term)
        comparison['%s_%s' % (old_eval, new_eval)].append((idx, grounding,
                                                           term))
    return comparison


def get_comparison_delta(groundings, c1, c2):
    def find_grounding(c, idx):
        for k, v in c.items():
            for entry in v:
                if entry[0] == idx:
                    return k, entry

    for idx, grounding in enumerate(groundings):
        c1g = find_grounding(c1, idx)
        c2g = find_grounding(c2, idx)
        if c1g[0] != c2g[0]:
            print(c1g, c2g)


# We now calculate various summary statistics and then print them
def get_sum_start(d, s):
    return sum([len(v) for k, v in d.items() if k.startswith(s)])


def get_sum_end(d, s):
    return sum([len(v) for k, v in d.items() if k.endswith(s)])


def print_statistics(comparison):
    old_correct = get_sum_start(comparison, 'correct')
    old_incorrect = get_sum_start(comparison, 'incorrect')
    old_ungrounded = get_sum_start(comparison, 'ungrounded')
    correct = get_sum_end(comparison, '_correct')
    incorrect = get_sum_end(comparison, '_incorrect')
    ungrounded = get_sum_end(comparison, '_ungrounded')
    unknown = get_sum_end(comparison, '_unknown')

    assert old_correct + old_incorrect + old_ungrounded == 300

    # Note: it can happen that a change in Gilda's behavior results in
    # producing a grounding that is not curated and is therefore of
    # "unknown" status. To cover this possibility, we calculate a lower
    # and an upper bound for these values
    prec = (correct / (correct + incorrect + unknown),
            (correct + unknown) / (correct + incorrect + unknown))
    recall = (correct / (correct + ungrounded),
              (correct + unknown) / (correct + unknown + ungrounded))
    fscore = (2 * prec[0] * recall[0] / (prec[0] + recall[0]),
              2 * prec[1] * recall[1] / (prec[1] + recall[1]))

    old_prec = old_correct / (old_correct + old_incorrect)
    old_recall = old_correct / (old_correct + old_ungrounded)
    old_fscore = 2 * old_prec * old_recall / (old_prec + old_recall)

    for k, v in comparison.items():
        print('%s: %s' % (k, len(v)))

    print('The reference statistics were:')
    print('- Precision: %.3f\n- Recall: %.3f\n- F-score: %.3f' %
          (old_prec, old_recall, old_fscore))
    print('The current statistics with Gilda are:')
    if unknown:
        print('- Precision: (%.3f, %.3f)\n- Recall: (%.3f, %.3f)'
              '\n- F-score: (%.3f, %.3f)' %
              tuple(itertools.chain(prec, recall, fscore)))
    else:
        print('- Precision: %.3f\n- Recall: %.3f\n- F-score: %.3f' %
              (prec[0], recall[0], fscore[0]))

    print()

    rows = [[old_prec, old_recall, old_fscore]]
    rows.extend(zip(prec, recall, fscore) if unknown \
                else [[prec[0], recall[0], fscore[0]]])
    df2 = pandas.DataFrame(
        rows,
        columns=["Precision", "Recall", "F_1"],
        index=["Reference", "Current_1", "Current_2"] if unknown else \
            ["Reference", "Current"],
    ).round(3)
    df2.columns.name = 'Trial'
    print(df2.to_latex(caption="FamPlex benchmarking results",
                       label='tab:famplex-benchmark-results'))

    df1 = pandas.DataFrame(
        [
            (*(x.capitalize() for x in k.split('_')), len(v))
            for k, v in comparison.items()
        ],
        columns=["Expected", "Actual", "Count"],
    )
    df1 = df1.pivot(index=["Expected"], columns=["Actual"], values=["Count"])
    print(df1.to_latex(caption="FamPlex benchmarking confusion matrix",
                       label="tab:famplex-confusion"))


def run_comparison(use_disamb=True):
    df = pandas.read_csv(url)
    groundings = process_fplx_groundings(df)
    comparison = make_comparison(groundings, use_disamb)
    print_statistics(comparison)
    return comparison


if __name__ == '__main__':
    comparison = run_comparison(use_disamb=True)



"""Module containing various string processing functions used for grounding."""
from typing import List, Tuple

import regex as re
import unidecode

from .greek_alphabet import greek_alphabet, greek_to_latin


#: A list of all kinds of dashes
dashes = [chr(0x2212), chr(0x002d)] + [chr(c) for c in range(0x2010, 0x2016)]


def replace_dashes(s, rep='-'):
    """Replace all types of dashes in a given string with a given replacement.

    Parameters
    ----------
    s : str
        The string in which all types of dashes should be replaced.
    rep : Optional[str]
        The string with which dashes should be replaced. By default, the plain
        ASCII dash (-) is used.

    Returns
    -------
    str
        The string in which dashes have been replaced.
    """
    for d in dashes:
        s = s.replace(d, rep)
    return s


def remove_dashes(s):
    """Remove all types of dashes in the given string.

    Parameters
    ----------
    s : str
        The string in which all types of dashes should be replaced.

    Returns
    -------
    str
        The string from which dashes have been removed.
    """
    return replace_dashes(s, '')


def replace_whitespace(s, rep=' '):
    """Replace any length white spaces in the given string with a replacement.

    Parameters
    ----------
    s : str
        The string in which any length whitespaces should be replaced.
    rep : Optional[str]
        The string with which all whitespace should be replaced. By default,
        the plain ASCII space ( ) is used.

    Returns
    -------
    str
        The string in which whitespaces have been replaced.
    """
    s = re.sub(r'\s+', rep, s)
    return s


def normalize(s):
    """Normalize white spaces, dashes and case of a given string.

    Parameters
    ----------
    s : str
        The string to be normalized.

    Returns
    -------
    str
        The normalized string.
    """
    s = replace_whitespace(s)
    s = remove_dashes(s)
    s = replace_unicode(s)
    s = s.lower()
    return s


def split_preserve_tokens(s):
    """Return split words of a string including the non-word tokens.

    Parameters
    ----------
    s : str
        The string to be split.

    Returns
    -------
    list of str
        The list of words in the string including the separator tokens,
        typically spaces and dashes..
    """
    return re.split(r'(\W)', s)


def replace_greek_uni(s):
    """Replace Greek spelled out letters with their unicode character."""
    for greek_uni, greek_spelled_out in greek_alphabet.items():
        s = s.replace(greek_spelled_out, greek_uni)
    return s


def replace_greek_latin(s):
    """Replace Greek spelled out letters with their latin character."""
    for greek_spelled_out, latin in greek_to_latin.items():
        s = s.replace(greek_spelled_out, latin)
    return s


def replace_greek_spelled_out(s):
    """Replace Greek unicode character with latin spelled out.
    """
    for greek_uni, greek_spelled_out in greek_alphabet.items():
        s = s.replace(greek_uni, greek_spelled_out)
    return s


def replace_unicode(s):
    """Replace unicode with ASCII equivalent, except Greek letters.

    Greek letters are handled separately and aren't replaced in this context.
    """
    if unidecode.unidecode(s) == s:
        return s
    return ''.join(unidecode.unidecode(c) if c not in greek_alphabet else c
                   for c in s)


def get_capitalization_pattern(word, beginning_of_sentence=False):
    """Return the type of capitalization for the string.

    Parameters
    ----------
    word : str
        The word whose capitalization is determined.
    beginning_of_sentence : Optional[bool]
        True if the word appears at the beginning of a sentence. Default: False

    Returns
    -------
    str
        The capitalization pattern of the given word. Returns one of the
        following: sentence_initial_cap, single_cap_letter, all_caps, all_lower,
        initial_cap, mixed.

    """
    if beginning_of_sentence and re.match(r'^\p{Lu}\p{Ll}*$', word):
        return 'sentence_initial_cap'
    elif re.match(r'^\p{Lu}$', word):
        return 'single_cap_letter'
    elif re.match(r'^\p{Lu}+$', word):
        return 'all_caps'
    elif re.match(r'^\p{Ll}+$', word):
        return 'all_lower'
    elif re.match(r'^\p{Lu}\p{Ll}+$', word):
        return 'initial_cap'
    else:
        return 'mixed'


def depluralize(word: str) -> List[Tuple[str, str]]:
    """Return the depluralized version of the word, along with a status flag.

    Parameters
    ----------
    word : str
        The word which is to be depluralized.

    Returns
    -------
    list of str pairs:
        The original word, if it is detected to be non-plural, or the
        depluralized version of the word, and a status flag representing the
        detected pluralization status of the
        word, with non_plural (e.g., BRAF), plural_oes (e.g., mosquitoes),
        plural_ies (e.g., antibodies), plural_es (e.g., switches),
        plural_cap_s (e.g., MAPKs), and plural_s (e.g., receptors).
    """
    # If the word doesn't end in s, we assume it's not plural
    if not word.endswith('s'):
        return [(word, 'non_plural')]
    # Another case is words ending in -sis (e.g., apoptosis), these are almost
    # exclusively non plural so we return here too
    elif word.endswith('sis'):
        return [(word, 'non_plural')]
    # This is the case when the word ends with an o which is pluralized as oes
    # e.g., mosquitoes
    elif word.endswith('oes'):
        return [(word[:-2], 'plural_oes'),
                (word[:-1], 'plural_s')]
    # This is the case when the word ends with a y which is pluralized as ies,
    # e.g., antibodies
    elif word.endswith('ies'):
        return [(word[:-3] + 'y', 'plural_ies'),
                (word[:-1], 'plural_s')]
    # These are the cases where words form plurals by adding -es so we
    # return by stripping it off. However, it's not possible to determine
    # if the word doesn't end in e.g., -xe or -se in a singluar form, and
    # so we also return a variant to account for this.
    elif word.endswith(('xes', 'ses', 'ches', 'shes')):
        return [(word[:-2], 'plural_es'), (word[:-1], 'plural_s')]
    # If the word is all caps and the last letter is an s, then it's a very
    # strong signal that it is pluralized so we have a custom return value
    # for that
    elif re.match(r'^\p{Lu}+$', word[:-1]):
        return [(word[:-1], 'plural_caps_s')]
    # Otherwise, we just go with the assumption that the last s is the
    # plural marker
    else:
        return [(word[:-1], 'plural_s')]
    # Note: there don't seem to be any compelling examples of -f or -fe -> ves
    # so it is not implemented


def replace_roman_arabic(s):
    match = roman_arabic_prefilter.match(s)
    if not match:
        return s
    else:
        pattern = roman_arabic_patterns.get(match.groups()[0].upper())
        return pattern[0].sub(pattern[1], s) if pattern else s


def _make_roman_arabic_patterns():
    roman_arabic = {
        'I': '1',
        'II': '2',
        'III': '3',
        'IV': '4',
        'V': '5',
        'VI': '6',
        'VII': '7',
        'VIII': '8',
        'IX': '9',
        'X': '10'
    }

    roman_arabic_patterns = {}
    for r, a in roman_arabic.items():
        for a, b in [(r, a), (a, r)]:
            roman_arabic_patterns[a] = (re.compile(r'^(.*[- ])(%s)$' % a,
                                                  re.IGNORECASE),
                                        r'\g<1>%s' % b)
    return roman_arabic_patterns


roman_arabic_patterns = _make_roman_arabic_patterns()
roman_arabic_prefilter = re.compile(r'^.*[- ](\d+|[IXV]+)$', re.IGNORECASE)



import os
import json

from bioid_evaluation import BioIDBenchmarker


HERE = os.path.dirname(os.path.abspath(__file__))
data_path = os.path.join(HERE, 'data')
try:
    with open(os.path.join(data_path, 'equivalences.json')) as f:
        equivalences = json.load(f)
except FileNotFoundError:
    equivalences = {}

benchmarker = BioIDBenchmarker(equivalences=equivalences)
benchmarker.ground_entities_with_gilda(context=True)

df = benchmarker.processed_data
df['top_grounding_no_context'] = df.groundings_no_context.\
    apply(lambda x: x[0][0] if x else None)

fp = \
    df[
        (~df.top_correct_no_context) &
        (
            df.top_grounding_no_context.apply(
                lambda x: x is not None and x.startswith('HGNC'))
        )
    ]

tp = \
    df[
        (df.top_correct_no_context) &
        (
            df.top_grounding_no_context.apply(
                lambda x: x is not None and x.startswith('HGNC')
            )
        )
    ]

pos = df[
    df.obj_synonyms.apply(lambda x: any([y.startswith('HGNC') for y in x]))
]


pr = len(tp) / (len(tp) + len(fp))
rc = len(tp) / len(pos)
f1 = 2/(1/pr + 1/rc)


df['top_grounding'] = df.groundings.\
    apply(lambda x: x[0][0] if x else None)

fp_disamb = \
    df[
        (~df.top_correct) &
        (
            df.top_grounding.apply(
                lambda x: x is not None and x.startswith('HGNC'))
        )
    ]

tp_disamb = \
    df[
        (df.top_correct) &
        (
            df.top_grounding.apply(
                lambda x: x is not None and x.startswith('HGNC')
            )
        )
    ]

pos_disamb = df[
    df.obj_synonyms.apply(lambda x: any([y.startswith('HGNC') for y in x]))
]


pr_disamb = len(tp_disamb) / (len(tp_disamb) + len(fp_disamb))
rc_disamb = len(tp_disamb) / len(pos_disamb)
f1_disamb = 2/(1/pr_disamb + 1/rc_disamb)



import os
import pandas
from collections import defaultdict
from gilda.generate_terms import *
from indra.databases import mesh_client


resources = os.path.join(os.path.dirname(__file__), os.path.pardir,
                         'gilda', 'resources')


def is_protein(mesh_id):
    return mesh_client.is_protein(mesh_id) or mesh_client.is_enzyme(mesh_id)


def is_chemical(mesh_id):
    return mesh_client.is_molecular(mesh_id)


def load_biomappings():
    """Load curated positive and negative mappings from Biomappings."""
    url_base = ('https://raw.githubusercontent.com/biopragmatics/biomappings/'
                'master/src/biomappings/resources/')
    positive_df = pandas.read_csv(url_base + 'mappings.tsv', sep='\t')
    negative_df = pandas.read_csv(url_base + 'incorrect.tsv', sep='\t')
    positive_mappings = defaultdict(list)
    negative_mappings = defaultdict(list)
    # These are the only relevant prefixes, there are mappings to
    # various other namespaces we don't need
    prefixes = {'fplx', 'chebi', 'go', 'hp', 'doid', 'efo', 'hgnc'}
    for mapping_df, mappings in ((positive_df, positive_mappings),
                                 (negative_df, negative_mappings)):
        for _, row in mapping_df.iterrows():
            # We only need exact matches.
            # TODO: should we consider non-exact matches to be effectively
            # negative?
            if row['relation'] != 'skos:exactMatch':
                continue
            # Look at both directions in which mesh mappings
            # can appear
            if row['source prefix'] == 'mesh':
                mesh_id = row['source identifier']
                other_ns = row['target prefix']
                other_id = row['target identifier']
            elif row['target prefix'] == 'mesh':
                mesh_id = row['target identifier']
                other_ns = row['source prefix']
                other_id = row['source identifier']
            else:
                continue
            if other_ns not in prefixes:
                continue
            # We make the namespace upper to be consistent
            # with Gilda
            mappings[mesh_id].append((other_ns.upper(), other_id))
    return positive_mappings, negative_mappings


def get_nonambiguous(maps):
    # If there is more than one mapping from MESH
    if len(maps) > 1:
        # We see if there are any name-level matches
        name_matches = [(me, te) for me, te in maps
                        if (me.entry_name.lower() if me.entry_name else '')
                            == (te.entry_name.lower() if te.entry_name else '')
                        # Corner case where we have multiple MeSH-based terms
                        # due to an orignal term from e.g., DOID having been
                        # mapped to MeSH
                        and me.db != te.db]

        # If we still have ambiguity, we print to the user
        if not name_matches or len(name_matches) > 1:
            return None, maps
        # Otherwise. we add the single name matches mapping
        else:
            return name_matches[0], []
    # If we map to only one thing, we keep that mapping
    else:
        return list(maps)[0], []


def resolve_duplicates(mappings):
    keep_mappings = []
    all_ambigs = []
    # First we deal with mappings from MESH
    for key, maps in mappings.items():
        maps_list = maps.values()
        keep, ambigs = get_nonambiguous(maps_list)
        if keep:
            keep_mappings.append(keep)
        if ambigs:
            all_ambigs += ambigs

    # Next we deal with mappings to MESH
    reverse_mappings = defaultdict(list)
    for mesh_term, other_term in keep_mappings:
        reverse_mappings[(other_term.db, other_term.id)].append((mesh_term,
                                                                 other_term))
    keep_mappings = []
    for maps in reverse_mappings.values():
        keep, ambigs = get_nonambiguous(maps)
        if keep:
            keep_mappings.append(keep)
        if ambigs:
            all_ambigs += ambigs

    return keep_mappings, all_ambigs


def dump_mappings(mappings, fname):
    def render_row(me, te):
        return '\t'.join([me.db, me.id, me.entry_name,
                          te.db, te.id, te.entry_name])

    with open(fname, 'w') as fh:
        for mesh_term, other_term in sorted(mappings, key=lambda x: x[0].id):
            # Corner case where we have multiple MeSH-based terms
            # due to an orignal term from e.g., DOID having been
            # mapped to MeSH
            if other_term.db != 'MESH':
                fh.write(render_row(mesh_term, other_term) + '\n')


def get_ambigs_by_db(ambigs):
    ambigs_by_db = defaultdict(list)
    for term in ambigs:
        ambigs_by_db[term.db].append(term)
    return dict(ambigs_by_db)


def get_mesh_mappings(ambigs):
    mappings_by_mesh_id = defaultdict(dict)
    for text, ambig_terms in ambigs.items():
        ambigs_by_db = get_ambigs_by_db(ambig_terms)
        #print('Considering %s' % text)
        #for term in ambig_terms:
        #    print('%s:%s %s' % (term.db, term.id, term.entry_name))
        order = [('FPLX', is_protein),
                 ('HGNC', is_protein),
                 ('CHEBI', is_chemical),
                 ('GO', lambda x: True),
                 ('DOID', lambda x: True),
                 ('HP', lambda x: True),
                 ('EFO', lambda x: True)]
        me = ambigs_by_db['MESH'][0]
        for ns, mesh_constraint in order:
            if len(ambigs_by_db.get(ns, [])) == 1 and mesh_constraint(me.id):
                mappings_by_mesh_id[me.id][(ambigs_by_db[ns][0].db,
                                            ambigs_by_db[ns][0].id)] = \
                        [me, ambigs_by_db[ns][0]]
                #print('Adding mapping for %s' % ns)
                break
        #print('--------------')
    return dict(mappings_by_mesh_id)


def find_ambiguities(terms, match_attr='text'):
    match_fun = lambda x: x.text if match_attr == 'text' else x.norm_text
    ambig_entries = defaultdict(list)
    for term in terms:
        # We consider it an ambiguity if the same text entry appears
        # multiple times
        ambig_entries[match_fun(term)].append(term)
    # There is a corner case where the match_fun matches two different
    # synonyms / variants of the same entry from the same database which
    # are not really considered ambiguity but need to be reduced to a single
    # entry to avoid being inadvertently filtered out later
    ambig_entries = {
        # Here, we make sure we only keep a single term with a given db and id
        norm_term: list({(term.db, term.id): term for term in matching_terms}.values())
        for norm_term, matching_terms in ambig_entries.items()
    }
    # It's only an ambiguity if there are two entries at least
    ambig_entries = {norm_term: matching_terms
                     for norm_term, matching_terms
                     in ambig_entries.items() if len(matching_terms) >= 2}
    # We filter out any ambiguities that contain not exactly one MeSH term
    ambig_entries = {k: v for k, v in ambig_entries.items()
                     if len([e for e in v if e.db == 'MESH']) == 1}
    print('Found a total of %d relevant ambiguities' % len(ambig_entries))
    return dict(ambig_entries)


def get_terms():
    terms = generate_mesh_terms(ignore_mappings=True) + \
        generate_go_terms() + \
        generate_hgnc_terms() + \
        generate_famplex_terms(ignore_mappings=True) + \
        generate_uniprot_terms(download=False) + \
        generate_chebi_terms() + \
        generate_efo_terms(ignore_mappings=True) + \
        generate_hp_terms(ignore_mappings=True) + \
        generate_doid_terms(ignore_mappings=True)
    terms = filter_out_duplicates(terms)
    return terms


def manual_go_mappings(terms_by_id_tuple):
    # Migrated from FamPlex and INDRA
    map = [
        ('D002465', 'GO:0048870'),
        ('D002914', 'GO:0042627'),
        ('D012374', 'GO:0120200'),
        ('D014158', 'GO:0006351'),
        ('D014176', 'GO:0006412'),
        ('D018919', 'GO:0001525'),
        ('D048708', 'GO:0016049'),
        ('D058750', 'GO:0001837'),
        ('D059767', 'GO:0000725')
    ]
    mappings_by_mesh_id = defaultdict(dict)
    for mid, gid in map:
        mt = terms_by_id_tuple[('MESH', mid)]
        gt = terms_by_id_tuple[('GO', gid)]
        mappings_by_mesh_id[mid][('GO', gid)] = (mt, gt)
    return dict(mappings_by_mesh_id)


if __name__ == '__main__':
    terms = get_terms()
    # We create a lookup of term objects by their db/id tuple
    # for quick lookups. We also add source db/ids here
    # because they can be relevant when finding terms for
    # Biomappings curations. Note that when loading e.g.,
    # DOID terms, the native xrefs from DOID to MESH
    # are applied, even if terms are loaded with the ignore_mappings
    # option which just turns of loading the mappings that are
    # generated in this script.
    known_mappings = set()
    terms_by_id_tuple = {}
    for term in terms:
        terms_by_id_tuple[(term.db, term.id)] = term
        if term.source_id:
            terms_by_id_tuple[(term.source_db, term.source_id)] = term
            known_mappings.add((term.db, term.id, term.source_db, term.source_id))
    # General ambiguities
    ambigs = find_ambiguities(terms, match_attr='text')
    mappings = get_mesh_mappings(ambigs)
    # Ambiguities that involve long strings but we allow normalized matches
    ambigs2 = find_ambiguities(terms, match_attr='norm_text')
    ambigs2 = {k: v for k, v in ambigs2.items() if len(k) > 6}
    mappings2 = get_mesh_mappings(ambigs2)
    for k, v in mappings2.items():
        if k not in mappings:
            mappings[k] = v
    # Mappings from GO terms
    mappings3 = manual_go_mappings(terms_by_id_tuple)
    for k, v in mappings3.items():
        if k not in mappings:
            mappings[k] = v

    # We now have to account for Biomappings curations
    positive_biomappings, negative_biomappings = load_biomappings()
    keys_to_remove = set()
    # Iterate over all the automatically proposed mappings
    for mesh_id, local_mappings in mappings.items():
        # If we already have a positive curation for the given MeSH ID
        # we want to replace the content automatically generated here
        # with the terms corresponding to the positive curation
        if mesh_id in positive_biomappings:
            other_ids = positive_biomappings[mesh_id]
            new_mappings = {}
            for other_id in other_ids:
                # If the other ID already exists, we just copy it over
                if other_id in mappings[mesh_id]:
                    new_mappings[other_id] = mappings[mesh_id][other_id]
                # If it doesn't exist yet, we look up a Term for it
                # and add it to the mappings
                else:
                    if other_id in terms_by_id_tuple:
                        mesh_term = terms_by_id_tuple[('MESH', mesh_id)]
                        other_term = terms_by_id_tuple[other_id]
                        new_mappings[other_id] = [mesh_term, other_term]
                    # This is a corner case where something is in Biomappings
                    # but not in the set of Gilda terms. This can happen
                    # if a term has been deprecated/replaced in an ontology.
                    # We ignore these mappings and just keep what we have.
                    else:
                        print('%s missing from set of terms' % str(other_id))
                        new_mappings = mappings[mesh_id]
            mappings[mesh_id] = new_mappings
        # If we have a negative curation for this MeSH ID, we make sure
        # that we remove any known incorrect mappings
        if mesh_id in negative_biomappings:
            other_ids = negative_biomappings[mesh_id]
            if mesh_id in mappings:
                for other_id in other_ids:
                    if other_id in mappings[mesh_id]:
                        mappings[mesh_id].pop(other_id, None)
                # If nothing left, we remove the whole MeSH ID key
                if not mappings[mesh_id]:
                    keys_to_remove.add(mesh_id)
    for key in keys_to_remove:
        mappings.pop(key)
    nonambig_mappings, ambig_mappings = resolve_duplicates(mappings)
    dump_mappings(nonambig_mappings, os.path.join(resources, 'mesh_mappings.tsv'))
    dump_mappings(ambig_mappings,
                  os.path.join(resources, 'mesh_ambig_mappings.tsv'))

    # Known mappings are useful for debugging
    #with open(os.path.join(resources, 'known_mappings.tsv'), 'w') as fh:
    #    for db, id, source_db, source_id in sorted(known_mappings):
    #        fh.write('\t'.join([db, id, source_db, source_id]) + '\n')